# Copyright 2025 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase

from portage.dbapi.bintree import binarytree


class BinarytreeBuildIdTestCase(TestCase):
    def testBinarytreeBuildId(self):
        cases = {
            "sec-keys/openpgp-keys-bzip2-20220406.gpkg.tar": -1,
            "sec-keys/openpgp-keys-bzip2/openpgp-keys-bzip2-20220406-1.gpkg.tar": 1,
            "sec-keys/openpgp-keys-bzip2-20220406.xpak": -1,
            "sec-keys/openpgp-keys-bzip2/openpgp-keys-bzip2-20220406-1.xpak": 1,
        }
        for filename, expected_build_id in cases.items():
            build_id = binarytree._parse_build_id(filename)
            self.assertEqual(
                build_id,
                expected_build_id,
                msg=f"Failed to parse build ID from '{filename}'",
            )



# Copyright 1999-2025 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from _emerge.AbstractEbuildProcess import AbstractEbuildProcess


class EbuildProcess(AbstractEbuildProcess):
    __slots__ = ("actionmap",)

    def _spawn(self, args, **kwargs):
        from portage.package.ebuild.doebuild import _doebuild_spawn, _spawn_actionmap

        actionmap = self.actionmap
        if actionmap is None:
            actionmap = _spawn_actionmap(self.settings)

        if self._dummy_pipe_fd is not None:
            self.settings["PORTAGE_PIPE_FD"] = str(self._dummy_pipe_fd)

        try:
            return _doebuild_spawn(
                self.phase, self.settings, actionmap=actionmap, **kwargs
            )
        finally:
            self.settings.pop("PORTAGE_PIPE_FD", None)



# Copyright 2015-2024 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2
#
# These constants are available from elfutils:
# https://sourceware.org/git/?p=elfutils.git;a=blob;f=libelf/elf.h;hb=HEAD

EI_CLASS = 4
ELFCLASS32 = 1
ELFCLASS64 = 2

EI_DATA = 5
ELFDATA2LSB = 1
ELFDATA2MSB = 2

E_TYPE = 16
ET_REL = 1
ET_EXEC = 2
ET_DYN = 3
ET_CORE = 4

E_MACHINE = 18
EM_SPARC = 2
EM_386 = 3
EM_68K = 4
EM_MIPS = 8
EM_PARISC = 15
EM_SPARC32PLUS = 18
EM_PPC = 20
EM_PPC64 = 21
EM_S390 = 22
EM_ARM = 40
EM_SH = 42
EM_SPARCV9 = 43
EM_ARC = 45
EM_IA_64 = 50
EM_X86_64 = 62
EM_ARC_COMPACT = 93
EM_ALTERA_NIOS2 = 113
EM_AARCH64 = 183
EM_ARC_COMPACT2 = 195
EM_AMDGPU = 224
EM_RISCV = 243
EM_BPF = 247
EM_ARC_COMPACT3_64 = 253
EM_ARC_COMPACT3 = 255
EM_LOONGARCH = 258
EM_ALPHA = 0x9026

E_ENTRY = 24
EF_MIPS_ABI = 0x0000F000
EF_MIPS_ABI2 = 0x00000020
E_MIPS_ABI_O32 = 0x00001000
E_MIPS_ABI_O64 = 0x00002000
E_MIPS_ABI_EABI32 = 0x00003000
E_MIPS_ABI_EABI64 = 0x00004000

EF_RISCV_RVC = 0x0001
EF_RISCV_FLOAT_ABI = 0x0006
EF_RISCV_FLOAT_ABI_SOFT = 0x0000
EF_RISCV_FLOAT_ABI_SINGLE = 0x0002
EF_RISCV_FLOAT_ABI_DOUBLE = 0x0004
EF_RISCV_FLOAT_ABI_QUAD = 0x0006

EF_LOONGARCH_ABI_LP64_SOFT_FLOAT = 0b001
EF_LOONGARCH_ABI_LP64_SINGLE_FLOAT = 0b010
EF_LOONGARCH_ABI_LP64_DOUBLE_FLOAT = 0b011
EF_LOONGARCH_ABI_ILP32_SOFT_FLOAT = 0b101
EF_LOONGARCH_ABI_ILP32_SINGLE_FLOAT = 0b110
EF_LOONGARCH_ABI_ILP32_DOUBLE_FLOAT = 0b111
EF_LOONGARCH_ABI_MASK = 0x07



# SPDX-License-Identifier: Apache-2.0
# Copyright 2012-2021 The Meson development team
# Copyright © 2021-2025 Intel Corporation
from __future__ import annotations

import collections
import enum
import functools
import os
import itertools
import typing as T

from .. import build
from .. import dependencies
from .. import options
from .. import mesonlib
from .. import mlog
from ..compilers import SUFFIX_TO_LANG, RunResult
from ..compilers.compilers import CompileCheckMode
from ..interpreterbase import (ObjectHolder, noPosargs, noKwargs,
                               FeatureNew, FeatureNewKwargs, disablerIfNotFound,
                               InterpreterException, InterpreterObject)
from ..interpreterbase.decorators import ContainerTypeInfo, typed_kwargs, KwargInfo, typed_pos_args
from ..options import OptionKey
from .interpreterobjects import (extract_required_kwarg, extract_search_dirs)
from .type_checking import REQUIRED_KW, in_set_validator, NoneType

if T.TYPE_CHECKING:
    from ..interpreter import Interpreter
    from ..compilers import Compiler
    from ..interpreterbase import TYPE_var, TYPE_kwargs
    from .kwargs import ExtractRequired, ExtractSearchDirs
    from .interpreter import SourceOutputs
    from ..mlog import TV_LoggableList

    from typing_extensions import TypedDict, Literal

    class GetSupportedArgumentKw(TypedDict):

        checked: Literal['warn', 'require', 'off']

    class AlignmentKw(TypedDict):

        prefix: str
        args: T.List[str]
        dependencies: T.List[dependencies.Dependency]

    class BaseCompileKW(TypedDict):
        no_builtin_args: bool
        include_directories: T.List[build.IncludeDirs]
        args: T.List[str]

    class CompileKW(BaseCompileKW, ExtractRequired):

        name: str
        dependencies: T.List[dependencies.Dependency]
        werror: bool

    class CommonKW(BaseCompileKW):

        prefix: str
        dependencies: T.List[dependencies.Dependency]

    class ComputeIntKW(CommonKW):

        guess: T.Optional[int]
        high: T.Optional[int]
        low: T.Optional[int]

    class HeaderKW(CommonKW, ExtractRequired):
        pass

    class HasKW(CommonKW, ExtractRequired):
        pass

    class HasArgumentKW(ExtractRequired):
        pass

    class FindLibraryKW(ExtractRequired, ExtractSearchDirs):

        disabler: bool
        has_headers: T.List[str]
        static: bool

        # This list must be all of the `HeaderKW` values with `header_`
        # prepended to the key
        header_args: T.List[str]
        header_dependencies: T.List[dependencies.Dependency]
        header_include_directories: T.List[build.IncludeDirs]
        header_no_builtin_args: bool
        header_prefix: str
        header_required: T.Union[bool, options.UserFeatureOption]

    class PreprocessKW(TypedDict):
        output: str
        compile_args: T.List[str]
        include_directories: T.List[build.IncludeDirs]
        dependencies: T.List[dependencies.Dependency]
        depends: T.List[T.Union[build.BuildTarget, build.CustomTarget, build.CustomTargetIndex]]


class _TestMode(enum.Enum):

    """Whether we're doing a compiler or linker check."""

    COMPILER = 0
    LINKER = 1


class TryRunResultHolder(ObjectHolder['RunResult']):
    def __init__(self, res: 'RunResult', interpreter: 'Interpreter'):
        super().__init__(res, interpreter)

    @noPosargs
    @noKwargs
    @InterpreterObject.method('returncode')
    def returncode_method(self, args: T.List['TYPE_var'], kwargs: 'TYPE_kwargs') -> int:
        return self.held_object.returncode

    @noPosargs
    @noKwargs
    @InterpreterObject.method('compiled')
    def compiled_method(self, args: T.List['TYPE_var'], kwargs: 'TYPE_kwargs') -> bool:
        return self.held_object.compiled

    @noPosargs
    @noKwargs
    @InterpreterObject.method('stdout')
    def stdout_method(self, args: T.List['TYPE_var'], kwargs: 'TYPE_kwargs') -> str:
        return self.held_object.stdout

    @noPosargs
    @noKwargs
    @InterpreterObject.method('stderr')
    def stderr_method(self, args: T.List['TYPE_var'], kwargs: 'TYPE_kwargs') -> str:
        return self.held_object.stderr


_ARGS_KW: KwargInfo[T.List[str]] = KwargInfo(
    'args',
    ContainerTypeInfo(list, str),
    listify=True,
    default=[],
)
_DEPENDENCIES_KW: KwargInfo[T.List['dependencies.Dependency']] = KwargInfo(
    'dependencies',
    ContainerTypeInfo(list, dependencies.Dependency),
    listify=True,
    default=[],
)
_DEPENDS_KW: KwargInfo[T.List[T.Union[build.BuildTarget, build.CustomTarget, build.CustomTargetIndex]]] = KwargInfo(
    'depends',
    ContainerTypeInfo(list, (build.BuildTarget, build.CustomTarget, build.CustomTargetIndex)),
    listify=True,
    default=[],
)
_INCLUDE_DIRS_KW: KwargInfo[T.List[build.IncludeDirs]] = KwargInfo(
    'include_directories',
    ContainerTypeInfo(list, build.IncludeDirs),
    default=[],
    listify=True,
)
_PREFIX_KW: KwargInfo[str] = KwargInfo(
    'prefix',
    (str, ContainerTypeInfo(list, str)),
    default='',
    since_values={list: '1.0.0'},
    convertor=lambda x: '\n'.join(x) if isinstance(x, list) else x)

_NO_BUILTIN_ARGS_KW = KwargInfo('no_builtin_args', bool, default=False)
_NAME_KW = KwargInfo('name', str, default='')
_WERROR_KW = KwargInfo('werror', bool, default=False, since='1.3.0')

# Many of the compiler methods take this kwarg signature exactly, this allows
# simplifying the `typed_kwargs` calls
_COMMON_KWS: T.List[KwargInfo] = [_ARGS_KW, _DEPENDENCIES_KW, _INCLUDE_DIRS_KW, _PREFIX_KW, _NO_BUILTIN_ARGS_KW]

# Common methods of compiles, links, runs, and similar
_COMPILES_KWS: T.List[KwargInfo] = [_NAME_KW, _ARGS_KW, _DEPENDENCIES_KW, _INCLUDE_DIRS_KW, _NO_BUILTIN_ARGS_KW,
                                    _WERROR_KW,
                                    REQUIRED_KW.evolve(since='1.5.0', default=False)]

_HEADER_KWS: T.List[KwargInfo] = [REQUIRED_KW.evolve(since='0.50.0', default=False), *_COMMON_KWS]
_HAS_REQUIRED_KW = REQUIRED_KW.evolve(since='1.3.0', default=False)

class CompilerHolder(ObjectHolder['Compiler']):
    preprocess_uid: T.Dict[str, itertools.count] = collections.defaultdict(itertools.count)

    def __init__(self, compiler: 'Compiler', interpreter: 'Interpreter'):
        super().__init__(compiler, interpreter)
        self.environment = self.env

    @property
    def compiler(self) -> 'Compiler':
        return self.held_object

    def _dep_msg(self, deps: T.List['dependencies.Dependency'], compile_only: bool, endl: str) -> str:
        msg_single = 'with dependency {}'
        msg_many = 'with dependencies {}'
        names = []
        for d in deps:
            if isinstance(d, dependencies.InternalDependency):
                FeatureNew.single_use('compiler method "dependencies" kwarg with internal dep', '0.57.0', self.subproject,
                                      location=self.current_node)
                continue
            if isinstance(d, dependencies.ExternalLibrary):
                if compile_only:
                    continue
                name = '-l' + d.name
            else:
                name = d.name
            names.append(name)
        if not names:
            return endl
        tpl = msg_many if len(names) > 1 else msg_single
        if endl is None:
            endl = ''
        return tpl.format(', '.join(names)) + endl

    @noPosargs
    @noKwargs
    @InterpreterObject.method('version')
    def version_method(self, args: T.List['TYPE_var'], kwargs: 'TYPE_kwargs') -> str:
        return self.compiler.version

    @noPosargs
    @noKwargs
    @InterpreterObject.method('cmd_array')
    def cmd_array_method(self, args: T.List['TYPE_var'], kwargs: 'TYPE_kwargs') -> T.List[str]:
        return self.compiler.exelist

    def _determine_args(self, kwargs: BaseCompileKW,
                        mode: CompileCheckMode = CompileCheckMode.LINK) -> T.List[str]:
        args: T.List[str] = []
        for i in kwargs['include_directories']:
            for idir in i.to_string_list(self.environment.get_source_dir(), self.environment.get_build_dir()):
                args.extend(self.compiler.get_include_args(idir, False))
        if not kwargs['no_builtin_args']:
            args += self.compiler.get_option_compile_args(None, self.interpreter.environment, self.subproject)
            args += self.compiler.get_option_std_args(None, self.interpreter.environment, self.subproject)
            if mode is CompileCheckMode.LINK:
                args.extend(self.compiler.get_option_link_args(None, self.interpreter.environment, self.subproject))
        if kwargs.get('werror', False):
            args.extend(self.compiler.get_werror_args())
        args.extend(kwargs['args'])
        return args

    def _determine_dependencies(self, deps: T.List['dependencies.Dependency'], compile_only: bool = False, endl: str = ':') -> T.Tuple[T.List['dependencies.Dependency'], str]:
        deps = dependencies.get_leaf_external_dependencies(deps)
        return deps, self._dep_msg(deps, compile_only, endl)

    @typed_pos_args('compiler.alignment', str)
    @typed_kwargs(
        'compiler.alignment',
        _PREFIX_KW,
        _ARGS_KW,
        _DEPENDENCIES_KW,
    )
    @InterpreterObject.method('alignment')
    def alignment_method(self, args: T.Tuple[str], kwargs: 'AlignmentKw') -> int:
        typename = args[0]
        deps, msg = self._determine_dependencies(kwargs['dependencies'], compile_only=self.compiler.is_cross)
        result, cached = self.compiler.alignment(typename, kwargs['prefix'], self.environment,
                                                 extra_args=kwargs['args'],
                                                 dependencies=deps)
        cached_msg = mlog.blue('(cached)') if cached else ''
        mlog.log('Checking for alignment of',
                 mlog.bold(typename, True), msg, mlog.bold(str(result)), cached_msg)
        return result

    @typed_pos_args('compiler.run', (str, mesonlib.File))
    @typed_kwargs('compiler.run', *_COMPILES_KWS)
    @InterpreterObject.method('run')
    def run_method(self, args: T.Tuple['mesonlib.FileOrString'], kwargs: 'CompileKW') -> 'RunResult':
        if self.compiler.language not in {'d', 'c', 'cpp', 'objc', 'objcpp', 'fortran'}:
            FeatureNew.single_use(f'compiler.run for {self.compiler.get_display_language()} language',
                                  '1.5.0', self.subproject, location=self.current_node)
        code = args[0]
        testname = kwargs['name']

        disabled, required, feature = extract_required_kwarg(kwargs, self.subproject, default=False)
        if disabled:
            if testname:
                mlog.log('Checking if', mlog.bold(testname, True), 'runs:', 'skipped: feature', mlog.bold(feature), 'disabled')
            return RunResult(compiled=True, returncode=0, stdout='', stderr='', cached=False)

        if isinstance(code, mesonlib.File):
            self.interpreter.add_build_def_file(code)
            code = mesonlib.File.from_absolute_file(
                code.rel_to_builddir(self.environment.source_dir))
        extra_args = functools.partial(self._determine_args, kwargs)
        deps, msg = self._determine_dependencies(kwargs['dependencies'], compile_only=False, endl=None)
        result = self.compiler.run(code, self.environment, extra_args=extra_args,
                                   dependencies=deps)
        if required and result.returncode != 0:
            raise InterpreterException(f'Could not run {testname if testname else "code"}')

        if testname:
            if not result.compiled:
                h = mlog.red('DID NOT COMPILE')
            elif result.returncode == 0:
                h = mlog.green('YES')
            else:
                h = mlog.red(f'NO ({result.returncode})')
            mlog.log('Checking if', mlog.bold(testname, True), msg, 'runs:', h)
        return result

    @noPosargs
    @noKwargs
    @InterpreterObject.method('get_id')
    def get_id_method(self, args: T.List['TYPE_var'], kwargs: 'TYPE_kwargs') -> str:
        return self.compiler.get_id()

    @noPosargs
    @noKwargs
    @FeatureNew('compiler.get_linker_id', '0.53.0')
    @InterpreterObject.method('get_linker_id')
    def get_linker_id_method(self, args: T.List['TYPE_var'], kwargs: 'TYPE_kwargs') -> str:
        return self.compiler.get_linker_id()

    @noPosargs
    @noKwargs
    @InterpreterObject.method('symbols_have_underscore_prefix')
    def symbols_have_underscore_prefix_method(self, args: T.List['TYPE_var'], kwargs: 'TYPE_kwargs') -> bool:
        '''
        Check if the compiler prefixes _ (underscore) to global C symbols
        See: https://en.wikipedia.org/wiki/Name_mangling#C
        '''
        return self.compiler.symbols_have_underscore_prefix(self.environment)

    @typed_pos_args('compiler.has_member', str, str)
    @typed_kwargs('compiler.has_member', _HAS_REQUIRED_KW, *_COMMON_KWS)
    @InterpreterObject.method('has_member')
    def has_member_method(self, args: T.Tuple[str, str], kwargs: 'HasKW') -> bool:
        typename, membername = args
        disabled, required, feature = extract_required_kwarg(kwargs, self.subproject, default=False)
        if disabled:
            mlog.log('Type', mlog.bold(typename, True), 'has member', mlog.bold(membername, True), 'skipped: feature', mlog.bold(feature), 'disabled')
            return False
        extra_args = functools.partial(self._determine_args, kwargs)
        deps, msg = self._determine_dependencies(kwargs['dependencies'])
        had, cached = self.compiler.has_members(typename, [membername], kwargs['prefix'],
                                                self.environment,
                                                extra_args=extra_args,
                                                dependencies=deps)
        cached_msg = mlog.blue('(cached)') if cached else ''
        if required and not had:
            raise InterpreterException(f'{self.compiler.get_display_language()} member {membername!r} of type {typename!r} not usable')
        elif had:
            hadtxt = mlog.green('YES')
        else:
            hadtxt = mlog.red('NO')
        mlog.log('Checking whether type', mlog.bold(typename, True),
                 'has member', mlog.bold(membername, True), msg, hadtxt, cached_msg)
        return had

    @typed_pos_args('compiler.has_members', str, varargs=str, min_varargs=1)
    @typed_kwargs('compiler.has_members', _HAS_REQUIRED_KW, *_COMMON_KWS)
    @InterpreterObject.method('has_members')
    def has_members_method(self, args: T.Tuple[str, T.List[str]], kwargs: 'HasKW') -> bool:
        typename, membernames = args
        members = mlog.bold(', '.join([f'"{m}"' for m in membernames]))
        disabled, required, feature = extract_required_kwarg(kwargs, self.subproject, default=False)
        if disabled:
            mlog.log('Type', mlog.bold(typename, True), 'has members', members, 'skipped: feature', mlog.bold(feature), 'disabled')
            return False
        extra_args = functools.partial(self._determine_args, kwargs)
        deps, msg = self._determine_dependencies(kwargs['dependencies'])
        had, cached = self.compiler.has_members(typename, membernames, kwargs['prefix'],
                                                self.environment,
                                                extra_args=extra_args,
                                                dependencies=deps)
        cached_msg = mlog.blue('(cached)') if cached else ''
        if required and not had:
            # print members as array: ['member1', 'member2']
            raise InterpreterException(f'{self.compiler.get_display_language()} members {membernames!r} of type {typename!r} not usable')
        elif had:
            hadtxt = mlog.green('YES')
        else:
            hadtxt = mlog.red('NO')
        mlog.log('Checking whether type', mlog.bold(typename, True),
                 'has members', members, msg, hadtxt, cached_msg)
        return had

    @typed_pos_args('compiler.has_function', str)
    @typed_kwargs('compiler.has_function', _HAS_REQUIRED_KW, *_COMMON_KWS)
    @InterpreterObject.method('has_function')
    def has_function_method(self, args: T.Tuple[str], kwargs: 'HasKW') -> bool:
        funcname = args[0]
        disabled, required, feature = extract_required_kwarg(kwargs, self.subproject, default=False)
        if disabled:
            mlog.log('Has function', mlog.bold(funcname, True), 'skipped: feature', mlog.bold(feature), 'disabled')
            return False
        extra_args = self._determine_args(kwargs)
        deps, msg = self._determine_dependencies(kwargs['dependencies'], compile_only=False)
        had, cached = self.compiler.has_function(funcname, kwargs['prefix'], self.environment,
                                                 extra_args=extra_args,
                                                 dependencies=deps)
        cached_msg = mlog.blue('(cached)') if cached else ''
        if required and not had:
            raise InterpreterException(f'{self.compiler.get_display_language()} function {funcname!r} not usable')
        elif had:
            hadtxt = mlog.green('YES')
        else:
            hadtxt = mlog.red('NO')
        mlog.log('Checking for function', mlog.bold(funcname, True), msg, hadtxt, cached_msg)
        return had

    @typed_pos_args('compiler.has_type', str)
    @typed_kwargs('compiler.has_type', _HAS_REQUIRED_KW, *_COMMON_KWS)
    @InterpreterObject.method('has_type')
    def has_type_method(self, args: T.Tuple[str], kwargs: 'HasKW') -> bool:
        typename = args[0]
        disabled, required, feature = extract_required_kwarg(kwargs, self.subproject, default=False)
        if disabled:
            mlog.log('Has type', mlog.bold(typename, True), 'skipped: feature', mlog.bold(feature), 'disabled')
            return False
        extra_args = functools.partial(self._determine_args, kwargs)
        deps, msg = self._determine_dependencies(kwargs['dependencies'])
        had, cached = self.compiler.has_type(typename, kwargs['prefix'], self.environment,
                                             extra_args=extra_args, dependencies=deps)
        cached_msg = mlog.blue('(cached)') if cached else ''
        if required and not had:
            raise InterpreterException(f'{self.compiler.get_display_language()} type {typename!r} not usable')
        elif had:
            hadtxt = mlog.green('YES')
        else:
            hadtxt = mlog.red('NO')
        mlog.log('Checking for type', mlog.bold(typename, True), msg, hadtxt, cached_msg)
        return had

    @FeatureNew('compiler.compute_int', '0.40.0')
    @typed_pos_args('compiler.compute_int', str)
    @typed_kwargs(
        'compiler.compute_int',
        KwargInfo('low', (int, NoneType)),
        KwargInfo('high', (int, NoneType)),
        KwargInfo('guess', (int, NoneType)),
        *_COMMON_KWS,
    )
    @InterpreterObject.method('compute_int')
    def compute_int_method(self, args: T.Tuple[str], kwargs: 'ComputeIntKW') -> int:
        expression = args[0]
        extra_args = functools.partial(self._determine_args, kwargs)
        deps, msg = self._determine_dependencies(kwargs['dependencies'], compile_only=self.compiler.is_cross)
        res = self.compiler.compute_int(expression, kwargs['low'], kwargs['high'],
                                        kwargs['guess'], kwargs['prefix'],
                                        self.environment, extra_args=extra_args,
                                        dependencies=deps)
        mlog.log('Computing int of', mlog.bold(expression, True), msg, res)
        return res

    @typed_pos_args('compiler.sizeof', str)
    @typed_kwargs('compiler.sizeof', *_COMMON_KWS)
    @InterpreterObject.method('sizeof')
    def sizeof_method(self, args: T.Tuple[str], kwargs: 'CommonKW') -> int:
        element = args[0]
        extra_args = functools.partial(self._determine_args, kwargs)
        deps, msg = self._determine_dependencies(kwargs['dependencies'], compile_only=self.compiler.is_cross)
        esize, cached = self.compiler.sizeof(element, kwargs['prefix'], self.environment,
                                             extra_args=extra_args, dependencies=deps)
        cached_msg = mlog.blue('(cached)') if cached else ''
        mlog.log('Checking for size of',
                 mlog.bold(element, True), msg, mlog.bold(str(esize)), cached_msg)
        return esize

    @FeatureNew('compiler.get_define', '0.40.0')
    @typed_pos_args('compiler.get_define', str)
    @typed_kwargs('compiler.get_define', *_COMMON_KWS)
    @InterpreterObject.method('get_define')
    def get_define_method(self, args: T.Tuple[str], kwargs: 'CommonKW') -> str:
        element = args[0]
        extra_args = functools.partial(self._determine_args, kwargs)
        deps, msg = self._determine_dependencies(kwargs['dependencies'])
        value, cached = self.compiler.get_define(element, kwargs['prefix'], self.environment,
                                                 extra_args=extra_args,
                                                 dependencies=deps)
        cached_msg = mlog.blue('(cached)') if cached else ''
        value_msg = '(undefined)' if value is None else value
        mlog.log('Fetching value of define', mlog.bold(element, True), msg, value_msg, cached_msg)
        return value if value is not None else ''

    @FeatureNew('compiler.has_define', '1.3.0')
    @typed_pos_args('compiler.has_define', str)
    @typed_kwargs('compiler.has_define', *_COMMON_KWS)
    @InterpreterObject.method('has_define')
    def has_define_method(self, args: T.Tuple[str], kwargs: 'CommonKW') -> bool:
        define_name = args[0]
        extra_args = functools.partial(self._determine_args, kwargs)
        deps, msg = self._determine_dependencies(kwargs['dependencies'], endl=None)
        value, cached = self.compiler.get_define(define_name, kwargs['prefix'], self.environment,
                                                 extra_args=extra_args,
                                                 dependencies=deps)
        cached_msg = mlog.blue('(cached)') if cached else ''
        h = mlog.green('YES') if value is not None else mlog.red('NO')
        mlog.log('Checking if define', mlog.bold(define_name, True), msg, 'exists:', h, cached_msg)

        return value is not None

    @typed_pos_args('compiler.compiles', (str, mesonlib.File))
    @typed_kwargs('compiler.compiles', *_COMPILES_KWS)
    @InterpreterObject.method('compiles')
    def compiles_method(self, args: T.Tuple['mesonlib.FileOrString'], kwargs: 'CompileKW') -> bool:
        code = args[0]
        testname = kwargs['name']

        disabled, required, feature = extract_required_kwarg(kwargs, self.subproject, default=False)
        if disabled:
            if testname:
                mlog.log('Checking if', mlog.bold(testname, True), 'compiles:', 'skipped: feature', mlog.bold(feature), 'disabled')
            return False

        if isinstance(code, mesonlib.File):
            if code.is_built:
                FeatureNew.single_use('compiler.compiles with file created at setup time', '1.2.0', self.subproject,
                                      'It was broken and either errored or returned false.', self.current_node)
            self.interpreter.add_build_def_file(code)
            code = mesonlib.File.from_absolute_file(
                code.absolute_path(self.environment.source_dir, self.environment.build_dir))
        extra_args = functools.partial(self._determine_args, kwargs)
        deps, msg = self._determine_dependencies(kwargs['dependencies'], endl=None)
        result, cached = self.compiler.compiles(code, self.environment,
                                                extra_args=extra_args,
                                                dependencies=deps)
        if required and not result:
            raise InterpreterException(f'Could not compile {testname}')

        if testname:
            if result:
                h = mlog.green('YES')
            else:
                h = mlog.red('NO')
            cached_msg = mlog.blue('(cached)') if cached else ''
            mlog.log('Checking if', mlog.bold(testname, True), msg, 'compiles:', h, cached_msg)
        return result

    @typed_pos_args('compiler.links', (str, mesonlib.File))
    @typed_kwargs('compiler.links', *_COMPILES_KWS)
    @InterpreterObject.method('links')
    def links_method(self, args: T.Tuple['mesonlib.FileOrString'], kwargs: 'CompileKW') -> bool:
        code = args[0]
        testname = kwargs['name']

        disabled, required, feature = extract_required_kwarg(kwargs, self.subproject, default=False)
        if disabled:
            if testname:
                mlog.log('Checking if', mlog.bold(testname, True), 'links:', 'skipped: feature', mlog.bold(feature), 'disabled')
            return False

        compiler = None
        if isinstance(code, mesonlib.File):
            if code.is_built:
                FeatureNew.single_use('compiler.links with file created at setup time', '1.2.0', self.subproject,
                                      'It was broken and either errored or returned false.', self.current_node)
            self.interpreter.add_build_def_file(code)
            code = mesonlib.File.from_absolute_file(
                code.absolute_path(self.environment.source_dir, self.environment.build_dir))
            suffix = code.suffix
            if suffix not in self.compiler.file_suffixes:
                for_machine = self.compiler.for_machine
                clist = self.interpreter.coredata.compilers[for_machine]
                if suffix not in SUFFIX_TO_LANG:
                    # just pass it to the compiler driver
                    mlog.warning(f'Unknown suffix for test file {code}')
                elif SUFFIX_TO_LANG[suffix] not in clist:
                    mlog.warning(f'Passed {SUFFIX_TO_LANG[suffix]} source to links method, not specified for {for_machine.get_lower_case_name()} machine.')
                else:
                    compiler = clist[SUFFIX_TO_LANG[suffix]]

        extra_args = functools.partial(self._determine_args, kwargs)
        deps, msg = self._determine_dependencies(kwargs['dependencies'], compile_only=False, endl=None)
        result, cached = self.compiler.links(code, self.environment,
                                             compiler=compiler,
                                             extra_args=extra_args,
                                             dependencies=deps)
        if required and not result:
            raise InterpreterException(f'Could not link {testname if testname else "code"}')

        if testname:
            if result:
                h = mlog.green('YES')
            else:
                h = mlog.red('NO')
            cached_msg = mlog.blue('(cached)') if cached else ''
            mlog.log('Checking if', mlog.bold(testname, True), msg, 'links:', h, cached_msg)
        return result

    @FeatureNew('compiler.check_header', '0.47.0')
    @typed_pos_args('compiler.check_header', str)
    @typed_kwargs('compiler.check_header', *_HEADER_KWS)
    @InterpreterObject.method('check_header')
    def check_header_method(self, args: T.Tuple[str], kwargs: 'HeaderKW') -> bool:
        hname = args[0]
        disabled, required, feature = extract_required_kwarg(kwargs, self.subproject, default=False)
        if disabled:
            mlog.log('Check usable header', mlog.bold(hname, True), 'skipped: feature', mlog.bold(feature), 'disabled')
            return False
        extra_args = functools.partial(self._determine_args, kwargs)
        deps, msg = self._determine_dependencies(kwargs['dependencies'])
        haz, cached = self.compiler.check_header(hname, kwargs['prefix'], self.environment,
                                                 extra_args=extra_args,
                                                 dependencies=deps)
        cached_msg = mlog.blue('(cached)') if cached else ''
        if required and not haz:
            raise InterpreterException(f'{self.compiler.get_display_language()} header {hname!r} not usable')
        elif haz:
            h = mlog.green('YES')
        else:
            h = mlog.red('NO')
        mlog.log('Check usable header', mlog.bold(hname, True), msg, h, cached_msg)
        return haz

    def _has_header_impl(self, hname: str, kwargs: 'HeaderKW') -> bool:
        disabled, required, feature = extract_required_kwarg(kwargs, self.subproject, default=False)
        if disabled:
            mlog.log('Has header', mlog.bold(hname, True), 'skipped: feature', mlog.bold(feature), 'disabled')
            return False
        extra_args = functools.partial(self._determine_args, kwargs)
        deps, msg = self._determine_dependencies(kwargs['dependencies'])
        haz, cached = self.compiler.has_header(hname, kwargs['prefix'], self.environment,
                                               extra_args=extra_args, dependencies=deps)
        cached_msg = mlog.blue('(cached)') if cached else ''
        if required and not haz:
            raise InterpreterException(f'{self.compiler.get_display_language()} header {hname!r} not found')
        elif haz:
            h = mlog.green('YES')
        else:
            h = mlog.red('NO')
        mlog.log('Has header', mlog.bold(hname, True), msg, h, cached_msg)
        return haz

    @typed_pos_args('compiler.has_header', str)
    @typed_kwargs('compiler.has_header', *_HEADER_KWS)
    @InterpreterObject.method('has_header')
    def has_header_method(self, args: T.Tuple[str], kwargs: 'HeaderKW') -> bool:
        return self._has_header_impl(args[0], kwargs)

    @typed_pos_args('compiler.has_header_symbol', str, str)
    @typed_kwargs('compiler.has_header_symbol', *_HEADER_KWS)
    @InterpreterObject.method('has_header_symbol')
    def has_header_symbol_method(self, args: T.Tuple[str, str], kwargs: 'HeaderKW') -> bool:
        hname, symbol = args
        disabled, required, feature = extract_required_kwarg(kwargs, self.subproject, default=False)
        if disabled:
            mlog.log('Header', mlog.bold(hname, True), 'has symbol', mlog.bold(symbol, True), 'skipped: feature', mlog.bold(feature), 'disabled')
            return False
        extra_args = functools.partial(self._determine_args, kwargs)
        deps, msg = self._determine_dependencies(kwargs['dependencies'])
        haz, cached = self.compiler.has_header_symbol(hname, symbol, kwargs['prefix'], self.environment,
                                                      extra_args=extra_args,
                                                      dependencies=deps)
        if required and not haz:
            raise InterpreterException(f'{self.compiler.get_display_language()} symbol {symbol} not found in header {hname}')
        elif haz:
            h = mlog.green('YES')
        else:
            h = mlog.red('NO')
        cached_msg = mlog.blue('(cached)') if cached else ''
        mlog.log('Header', mlog.bold(hname, True), 'has symbol', mlog.bold(symbol, True), msg, h, cached_msg)
        return haz

    def notfound_library(self, libname: str) -> 'dependencies.ExternalLibrary':
        lib = dependencies.ExternalLibrary(libname, None,
                                           self.environment,
                                           self.compiler.language,
                                           silent=True)
        return lib

    @disablerIfNotFound
    @typed_pos_args('compiler.find_library', str)
    @typed_kwargs(
        'compiler.find_library',
        KwargInfo('required', (bool, options.UserFeatureOption), default=True),
        KwargInfo('has_headers', ContainerTypeInfo(list, str), listify=True, default=[], since='0.50.0'),
        KwargInfo('static', (bool, NoneType), since='0.51.0'),
        KwargInfo('disabler', bool, default=False, since='0.49.0'),
        KwargInfo('dirs', ContainerTypeInfo(list, str), listify=True, default=[]),
        *(k.evolve(name=f'header_{k.name}') for k in _HEADER_KWS)
    )
    @InterpreterObject.method('find_library')
    def find_library_method(self, args: T.Tuple[str], kwargs: 'FindLibraryKW') -> 'dependencies.ExternalLibrary':
        # TODO add dependencies support?
        libname = args[0]

        disabled, required, feature = extract_required_kwarg(kwargs, self.subproject)
        if disabled:
            mlog.log('Library', mlog.bold(libname), 'skipped: feature', mlog.bold(feature), 'disabled')
            return self.notfound_library(libname)

        # This could be done with a comprehension, but that confuses the type
        # checker, and having it check this seems valuable
        has_header_kwargs: 'HeaderKW' = {
            'required': required,
            'args': kwargs['header_args'],
            'dependencies': kwargs['header_dependencies'],
            'include_directories': kwargs['header_include_directories'],
            'prefix': kwargs['header_prefix'],
            'no_builtin_args': kwargs['header_no_builtin_args'],
        }
        for h in kwargs['has_headers']:
            if not self._has_header_impl(h, has_header_kwargs):
                return self.notfound_library(libname)

        search_dirs = extract_search_dirs(kwargs)

        prefer_static = self.environment.coredata.optstore.get_value_for(OptionKey('prefer_static'))
        if kwargs['static'] is True:
            libtype = mesonlib.LibType.STATIC
        elif kwargs['static'] is False:
            libtype = mesonlib.LibType.SHARED
        elif prefer_static:
            libtype = mesonlib.LibType.PREFER_STATIC
        else:
            libtype = mesonlib.LibType.PREFER_SHARED
        linkargs = self.compiler.find_library(libname, self.environment, search_dirs, libtype)
        if required and not linkargs:
            if libtype == mesonlib.LibType.PREFER_SHARED:
                libtype_s = 'shared or static'
            else:
                libtype_s = libtype.name.lower()
            raise InterpreterException('{} {} library {!r} not found'
                                       .format(self.compiler.get_display_language(),
                                               libtype_s, libname))
        lib = dependencies.ExternalLibrary(libname, linkargs, self.environment,
                                           self.compiler.language)
        return lib

    def _has_argument_impl(self, arguments: T.Union[str, T.List[str]],
                           mode: _TestMode = _TestMode.COMPILER,
                           kwargs: T.Optional['ExtractRequired'] = None) -> bool:
        """Shared implementation for methods checking compiler and linker arguments."""
        # This simplifies the callers
        if isinstance(arguments, str):
            arguments = [arguments]
        logargs: TV_LoggableList = [
            'Compiler for',
            self.compiler.get_display_language(),
            'supports{}'.format(' link' if mode is _TestMode.LINKER else ''),
            'arguments {}:'.format(' '.join(arguments)),
        ]
        kwargs = kwargs or {'required': False}
        disabled, required, feature = extract_required_kwarg(kwargs, self.subproject, default=False)
        if disabled:
            logargs += ['skipped: feature', mlog.bold(feature), 'disabled']
            mlog.log(*logargs)
            return False
        test = self.compiler.has_multi_link_arguments if mode is _TestMode.LINKER else self.compiler.has_multi_arguments
        result, cached = test(arguments, self.environment)
        if required and not result:
            logargs += ['not usable']
            raise InterpreterException(*logargs)
        logargs += [
            mlog.green('YES') if result else mlog.red('NO'),
            mlog.blue('(cached)') if cached else '',
        ]
        mlog.log(*logargs)
        return result

    @typed_pos_args('compiler.has_argument', str)
    @typed_kwargs('compiler.has_argument', _HAS_REQUIRED_KW)
    @InterpreterObject.method('has_argument')
    def has_argument_method(self, args: T.Tuple[str], kwargs: 'HasArgumentKW') -> bool:
        return self._has_argument_impl([args[0]], kwargs=kwargs)

    @typed_pos_args('compiler.has_multi_arguments', varargs=str)
    @typed_kwargs('compiler.has_multi_arguments', _HAS_REQUIRED_KW)
    @FeatureNew('compiler.has_multi_arguments', '0.37.0')
    @InterpreterObject.method('has_multi_arguments')
    def has_multi_arguments_method(self, args: T.Tuple[T.List[str]], kwargs: 'HasArgumentKW') -> bool:
        return self._has_argument_impl(args[0], kwargs=kwargs)

    @FeatureNew('compiler.get_supported_arguments', '0.43.0')
    @typed_pos_args('compiler.get_supported_arguments', varargs=str)
    @typed_kwargs(
        'compiler.get_supported_arguments',
        KwargInfo('checked', str, default='off', since='0.59.0',
                  validator=in_set_validator({'warn', 'require', 'off'})),
    )
    @InterpreterObject.method('get_supported_arguments')
    def get_supported_arguments_method(self, args: T.Tuple[T.List[str]], kwargs: 'GetSupportedArgumentKw') -> T.List[str]:
        supported_args: T.List[str] = []
        checked = kwargs['checked']

        for arg in args[0]:
            if not self._has_argument_impl([arg]):
                msg = f'Compiler for {self.compiler.get_display_language()} does not support "{arg}"'
                if checked == 'warn':
                    mlog.warning(msg)
                elif checked == 'require':
                    raise mesonlib.MesonException(msg)
            else:
                supported_args.append(arg)
        return supported_args

    @noKwargs
    @typed_pos_args('compiler.first_supported_argument', varargs=str)
    @InterpreterObject.method('first_supported_argument')
    def first_supported_argument_method(self, args: T.Tuple[T.List[str]], kwargs: 'TYPE_kwargs') -> T.List[str]:
        for arg in args[0]:
            if self._has_argument_impl([arg]):
                mlog.log('First supported argument:', mlog.bold(arg))
                return [arg]
        mlog.log('First supported argument:', mlog.red('None'))
        return []

    @FeatureNew('compiler.has_link_argument', '0.46.0')
    @typed_pos_args('compiler.has_link_argument', str)
    @typed_kwargs('compiler.has_link_argument', _HAS_REQUIRED_KW)
    @InterpreterObject.method('has_link_argument')
    def has_link_argument_method(self, args: T.Tuple[str], kwargs: 'HasArgumentKW') -> bool:
        return self._has_argument_impl([args[0]], mode=_TestMode.LINKER, kwargs=kwargs)

    @FeatureNew('compiler.has_multi_link_argument', '0.46.0')
    @typed_pos_args('compiler.has_multi_link_argument', varargs=str)
    @typed_kwargs('compiler.has_multi_link_argument', _HAS_REQUIRED_KW)
    @InterpreterObject.method('has_multi_link_arguments')
    def has_multi_link_arguments_method(self, args: T.Tuple[T.List[str]], kwargs: 'HasArgumentKW') -> bool:
        return self._has_argument_impl(args[0], mode=_TestMode.LINKER, kwargs=kwargs)

    @FeatureNew('compiler.get_supported_link_arguments', '0.46.0')
    @noKwargs
    @typed_pos_args('compiler.get_supported_link_arguments', varargs=str)
    @InterpreterObject.method('get_supported_link_arguments')
    def get_supported_link_arguments_method(self, args: T.Tuple[T.List[str]], kwargs: 'TYPE_kwargs') -> T.List[str]:
        supported_args: T.List[str] = []
        for arg in args[0]:
            if self._has_argument_impl([arg], mode=_TestMode.LINKER):
                supported_args.append(arg)
        return supported_args

    @FeatureNew('compiler.first_supported_link_argument', '0.46.0')
    @noKwargs
    @typed_pos_args('compiler.first_supported_link_argument', varargs=str)
    @InterpreterObject.method('first_supported_link_argument')
    def first_supported_link_argument_method(self, args: T.Tuple[T.List[str]], kwargs: 'TYPE_kwargs') -> T.List[str]:
        for arg in args[0]:
            if self._has_argument_impl([arg], mode=_TestMode.LINKER):
                mlog.log('First supported link argument:', mlog.bold(arg))
                return [arg]
        mlog.log('First supported link argument:', mlog.red('None'))
        return []

    def _has_function_attribute_impl(self, attr: str, kwargs: T.Optional['ExtractRequired'] = None) -> bool:
        """Common helper for function attribute testing."""
        logargs: TV_LoggableList = [
            f'Compiler for {self.compiler.get_display_language()} supports function attribute {attr}:',
        ]
        kwargs = kwargs or {'required': False}
        disabled, required, feature = extract_required_kwarg(kwargs, self.subproject, default=False)
        if disabled:
            logargs += ['skipped: feature', mlog.bold(feature), 'disabled']
            mlog.log(*logargs)
            return False
        had, cached = self.compiler.has_func_attribute(attr, self.environment)
        if required and not had:
            logargs += ['not usable']
            raise InterpreterException(*logargs)
        logargs += [
            mlog.green('YES') if had else mlog.red('NO'),
            mlog.blue('(cached)') if cached else ''
        ]
        mlog.log(*logargs)
        return had

    @FeatureNew('compiler.has_function_attribute', '0.48.0')
    @typed_pos_args('compiler.has_function_attribute', str)
    @typed_kwargs('compiler.has_function_attribute', _HAS_REQUIRED_KW)
    @InterpreterObject.method('has_function_attribute')
    def has_func_attribute_method(self, args: T.Tuple[str], kwargs: 'HasArgumentKW') -> bool:
        return self._has_function_attribute_impl(args[0], kwargs)

    @FeatureNew('compiler.get_supported_function_attributes', '0.48.0')
    @noKwargs
    @typed_pos_args('compiler.get_supported_function_attributes', varargs=str)
    @InterpreterObject.method('get_supported_function_attributes')
    def get_supported_function_attributes_method(self, args: T.Tuple[T.List[str]], kwargs: 'TYPE_kwargs') -> T.List[str]:
        return [a for a in args[0] if self._has_function_attribute_impl(a)]

    @FeatureNew('compiler.get_argument_syntax', '0.49.0')
    @noPosargs
    @noKwargs
    @InterpreterObject.method('get_argument_syntax')
    def get_argument_syntax_method(self, args: T.List['TYPE_var'], kwargs: 'TYPE_kwargs') -> str:
        return self.compiler.get_argument_syntax()

    @FeatureNew('compiler.preprocess', '0.64.0')
    @FeatureNewKwargs('compiler.preprocess', '1.3.2', ['compile_args'], extra_message='compile_args were ignored before this version')
    @typed_pos_args('compiler.preprocess', varargs=(str, mesonlib.File, build.CustomTarget, build.CustomTargetIndex, build.GeneratedList), min_varargs=1)
    @typed_kwargs(
        'compiler.preprocess',
        KwargInfo('output', str, default='@PLAINNAME@.i'),
        KwargInfo('compile_args', ContainerTypeInfo(list, str), listify=True, default=[]),
        _INCLUDE_DIRS_KW,
        _DEPENDENCIES_KW.evolve(since='1.1.0'),
        _DEPENDS_KW.evolve(since='1.4.0'),
    )
    @InterpreterObject.method('preprocess')
    def preprocess_method(self, args: T.Tuple[T.List['mesonlib.FileOrString']], kwargs: 'PreprocessKW') -> T.List[build.CustomTargetIndex]:
        compiler = self.compiler.get_preprocessor()
        _sources: T.List[mesonlib.File] = self.interpreter.source_strings_to_files(args[0])
        sources = T.cast('T.List[SourceOutputs]', _sources)
        if any(isinstance(s, (build.CustomTarget, build.CustomTargetIndex, build.GeneratedList)) for s in sources):
            FeatureNew.single_use('compiler.preprocess with generated sources', '1.1.0', self.subproject,
                                  location=self.current_node)

        tg_counter = next(self.preprocess_uid[self.interpreter.subdir])
        if tg_counter > 0:
            FeatureNew.single_use('compiler.preprocess used multiple times', '1.1.0', self.subproject,
                                  location=self.current_node)
        tg_name = f'preprocessor_{tg_counter}'
        tg = build.CompileTarget(
            tg_name,
            self.interpreter.subdir,
            self.subproject,
            self.environment,
            sources,
            kwargs['output'],
            compiler,
            self.interpreter.backend,
            kwargs['compile_args'],
            kwargs['include_directories'],
            kwargs['dependencies'],
            kwargs['depends'])
        self.interpreter.add_target(tg.name, tg)
        # Expose this target as list of its outputs, so user can pass them to
        # other targets, list outputs, etc.
        private_dir = os.path.relpath(self.interpreter.backend.get_target_private_dir(tg), self.interpreter.subdir)
        return [build.CustomTargetIndex(tg, os.path.join(private_dir, o)) for o in tg.outputs]



# Copyright 2013-2023 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import functools

from portage.checksum import perform_multiple_checksums
from portage.util._async.AsyncFunction import AsyncFunction


class FileDigester(AsyncFunction):
    """
    Asynchronously generate file digests. Pass in file_path and
    hash_names, and after successful execution, the digests
    attribute will be a dict containing all of the requested
    digests.
    """

    __slots__ = (
        "file_path",
        "hash_names",
    )

    def _start(self):
        self.target = functools.partial(
            perform_multiple_checksums, self.file_path, hashes=self.hash_names
        )
        super()._start()

    @property
    def digests(self):
        return self.result



# Copyright 2025 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class VirtualCycleTestCase(TestCase):
    def testVirtualCycle(self):
        ebuilds = {
            "app-misc/foo-1": {
                "EAPI": "8",
                "RDEPEND": "virtual/A",
            },
            "virtual/A-1": {
                "EAPI": "8",
                "RDEPEND": "virtual/B",
            },
            "virtual/B-1": {
                "EAPI": "8",
                "RDEPEND": "virtual/C",
            },
            "virtual/C-1": {
                "EAPI": "8",
                "RDEPEND": "virtual/A",
            },
            "app-misc/bar-1": {
                "EAPI": "8",
                "RDEPEND": "virtual/gzip",
            },
            "virtual/gzip-1": {
                "EAPI": "8",
                "RDEPEND": "virtual/gzip",
            },
        }

        test_cases = (
            # Test direct virtual cycle for bug 965570.
            ResolverPlaygroundTestCase(
                ["app-misc/bar"],
                success=False,
                virtual_cycle={"virtual/gzip-1"},
            ),
            # Test indirect virtual cycle for bug 965570.
            ResolverPlaygroundTestCase(
                ["app-misc/foo"],
                success=False,
                virtual_cycle={"virtual/A-1", "virtual/B-1", "virtual/C-1"},
            ),
        )

        playground = ResolverPlayground(debug=False, ebuilds=ebuilds)

        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.debug = False
            playground.cleanup()



# SPDX-License-Identifier: Apache-2.0
# Copyright 2017 The Meson development team

"""Code that creates simple startup projects."""

from __future__ import annotations

from pathlib import Path
from enum import Enum
import subprocess
import shutil
import sys
import os
import re
from glob import glob
import typing as T

from mesonbuild import build, mesonlib, mlog
from mesonbuild.coredata import FORBIDDEN_TARGET_NAMES
from mesonbuild.environment import detect_ninja
from mesonbuild.templates.mesontemplates import create_meson_build
from mesonbuild.templates.samplefactory import sample_generator
from mesonbuild.options import OptionKey

if T.TYPE_CHECKING:
    import argparse

    from typing_extensions import Protocol, Literal

    class Arguments(Protocol):

        srcfiles: T.List[Path]
        wd: str
        name: str
        executable: str
        deps: str
        language: Literal['c', 'cpp', 'cs', 'cuda', 'd', 'fortran', 'java', 'rust', 'objc', 'objcpp', 'vala']
        build: bool
        builddir: str
        force: bool
        type: Literal['executable', 'library']
        version: str


FORTRAN_SUFFIXES = {'.f', '.for', '.F', '.f90', '.F90'}
LANG_SUFFIXES = {'.c', '.cc', '.cpp', '.cs', '.cu', '.d', '.m', '.mm', '.rs', '.java', '.vala'} | FORTRAN_SUFFIXES
LANG_SUPPORTED = {'c', 'cpp', 'cs', 'cuda', 'd', 'fortran', 'java', 'rust', 'objc', 'objcpp', 'vala'}

DEFAULT_PROJECT = 'executable'
DEFAULT_VERSION = '0.1'
class DEFAULT_TYPES(Enum):
    EXE = 'executable'
    LIB = 'library'

INFO_MESSAGE = '''Sample project created. To build it run the
following commands:

meson setup builddir
meson compile -C builddir
'''


def create_sample(options: Arguments) -> None:
    '''
    Based on what arguments are passed we check for a match in language
    then check for project type and create new Meson samples project.
    '''
    sample_gen = sample_generator(options)
    if options.type == DEFAULT_TYPES['EXE'].value:
        sample_gen.create_executable()
    elif options.type == DEFAULT_TYPES['LIB'].value:
        sample_gen.create_library()
    else:
        raise RuntimeError('Unreachable code')
    print(INFO_MESSAGE)

def autodetect_options(options: Arguments, sample: bool = False) -> None:
    '''
    Here we autodetect options for args not passed in so don't have to
    think about it.
    '''
    if not options.name:
        options.name = Path().resolve().stem
        if not re.match('[a-zA-Z_][a-zA-Z0-9]*', options.name) and sample:
            raise SystemExit(f'Name of current directory "{options.name}" is not usable as a sample project name.\n'
                             'Specify a project name with --name.')
        print(f'Using "{options.name}" (name of current directory) as project name.')
    if not options.executable:
        options.executable = options.name
        print(f'Using "{options.executable}" (project name) as name of executable to build.')
    if options.executable in FORBIDDEN_TARGET_NAMES:
        raise mesonlib.MesonException(f'Executable name {options.executable!r} is reserved for Meson internal use. '
                                      'Refusing to init an invalid project.')
    if sample:
        # The rest of the autodetection is not applicable to generating sample projects.
        return
    if not options.srcfiles:
        srcfiles: T.List[Path] = []
        for f in (f for f in Path().iterdir() if f.is_file()):
            if f.suffix in LANG_SUFFIXES:
                srcfiles.append(f)
        if not srcfiles:
            raise SystemExit('No recognizable source files found.\n'
                             'Run meson init in an empty directory to create a sample project.')
        options.srcfiles = srcfiles
        print("Detected source files: " + ' '.join(str(s) for s in srcfiles))
    if not options.language:
        for f in options.srcfiles:
            if f.suffix == '.c':
                options.language = 'c'
                break
            if f.suffix in {'.cc', '.cpp'}:
                options.language = 'cpp'
                break
            if f.suffix == '.cs':
                options.language = 'cs'
                break
            if f.suffix == '.cu':
                options.language = 'cuda'
                break
            if f.suffix == '.d':
                options.language = 'd'
                break
            if f.suffix in FORTRAN_SUFFIXES:
                options.language = 'fortran'
                break
            if f.suffix == '.rs':
                options.language = 'rust'
                break
            if f.suffix == '.m':
                options.language = 'objc'
                break
            if f.suffix == '.mm':
                options.language = 'objcpp'
                break
            if f.suffix == '.java':
                options.language = 'java'
                break
            if f.suffix == '.vala':
                options.language = 'vala'
                break
        if not options.language:
            raise SystemExit("Can't autodetect language, please specify it with -l.")
        print("Detected language: " + options.language)

# Note: when adding arguments, please also add them to the completion
# scripts in $MESONSRC/data/shell-completions/
def add_arguments(parser: 'argparse.ArgumentParser') -> None:
    '''
    Here we add args for that the user can passed when making a new
    Meson project.
    '''
    parser.add_argument("srcfiles", metavar="sourcefile", nargs="*", type=Path, help="source files. default: all recognized files in current directory")
    parser.add_argument('-C', dest='wd', action=mesonlib.RealPathAction,
                        help='directory to cd into before running')
    parser.add_argument("-n", "--name", help="project name. default: name of current directory")
    parser.add_argument("-e", "--executable", help="executable name. default: project name")
    parser.add_argument("-d", "--deps", help="dependencies, comma-separated")
    parser.add_argument("-l", "--language", choices=sorted(LANG_SUPPORTED), help="project language. default: autodetected based on source files")
    parser.add_argument("-b", "--build", action='store_true', help="build after generation")
    parser.add_argument("--builddir", default='build', help="directory for build")
    parser.add_argument("-f", "--force", action="store_true", help="force overwrite of existing files and directories.")
    parser.add_argument('--type', default=DEFAULT_PROJECT, choices=('executable', 'library'), help=f"project type. default: {DEFAULT_PROJECT} based project")
    parser.add_argument('--version', default=DEFAULT_VERSION, help=f"project version. default: {DEFAULT_VERSION}")

def run(options: Arguments) -> int:
    '''
    Here we generate the new Meson sample project.
    '''
    if not Path(options.wd).exists():
        sys.exit('Project source root directory not found. Run this command in source directory root.')
    os.chdir(options.wd)

    if not glob('*'):
        autodetect_options(options, sample=True)
        if not options.language:
            print('Defaulting to generating a C language project.')
            options.language = 'c'
        create_sample(options)
    else:
        autodetect_options(options)
        if Path('meson.build').is_file() and not options.force:
            raise SystemExit('meson.build already exists. Use --force to overwrite.')
        create_meson_build(options)
    if options.build:
        if Path(options.builddir).is_dir() and options.force:
            print('Build directory already exists, deleting it.')
            shutil.rmtree(options.builddir)
        print('Building...')
        cmd = mesonlib.get_meson_command() + ['setup', options.builddir]
        ret = subprocess.run(cmd)
        if ret.returncode:
            raise SystemExit

        b = build.load(options.builddir)
        need_vsenv = T.cast('bool', b.environment.coredata.optstore.get_value_for(OptionKey('vsenv')))
        vsenv_active = mesonlib.setup_vsenv(need_vsenv)
        if vsenv_active:
            mlog.log(mlog.green('INFO:'), 'automatically activated MSVC compiler environment')

        cmd = detect_ninja() + ['-C', options.builddir]
        ret = subprocess.run(cmd)
        if ret.returncode:
            raise SystemExit
    return 0



# (c) 2022-2025 Michał Górny
# SPDX-License-Identifier: GPL-2.0-or-later

import filecmp
import functools
import importlib.util
import os
import os.path
import typing

from pathlib import Path, PurePath

from gpep517.utils import DEFAULT_PREFIX, install_scheme_dict, logger


def install_wheel_impl(args, wheel: Path):
    from installer import install
    from installer.destinations import SchemeDictionaryDestination
    from installer.records import RecordEntry
    from installer.sources import WheelFile
    from installer.utils import Scheme, get_launcher_kind

    class DeduplicatingDestination(SchemeDictionaryDestination):
        def __init__(
            self,
            *args,
            overwrite: bool,
            symlink_pyc: bool,
            symlink_to: typing.Optional[PurePath],
            **kwargs,
        ) -> None:
            super().__init__(*args, **kwargs)

            self.overwrite = overwrite
            self.symlink_pyc = symlink_pyc
            self.symlink_to = symlink_to

            if symlink_to is not None:
                if os.name == "nt":
                    raise RuntimeError(
                        "--symlink-to is not supported on Windows")
                purelib_path = PurePath(self.scheme_dict["purelib"])
                if self.scheme_dict["platlib"] != self.scheme_dict["purelib"]:
                    raise NotImplementedError(
                        "The implementation currently requires that platlib "
                        f"({self.scheme_dict['platlib']!r}) is the same "
                        f"as purelib ({self.scheme_dict['purelib']!r})"
                    )
                self.destdir_purelib = (
                    Path(self.destdir) /
                    purelib_path.relative_to(purelib_path.anchor))

        @functools.cache
        def check_symlink_to(self) -> None:
            full_symlink_to = self.destdir_purelib / self.symlink_to
            if not full_symlink_to.exists():
                raise FileNotFoundError(
                    "--symlink-to references a path that does not exist "
                    f"in --destdir: {full_symlink_to}"
                )

        def write_to_fs(
            self,
            scheme: Scheme,
            path: str,
            stream: typing.BinaryIO,
            is_executable: bool,
        ) -> RecordEntry:
            try:
                ret = super().write_to_fs(scheme, path, stream, is_executable)
            except FileExistsError:
                if not self.overwrite:
                    raise
                relative_path = PurePath(self.scheme_dict[scheme]) / path
                full_path = Path(self.destdir).joinpath(
                    relative_path.relative_to(relative_path.anchor))
                full_path.unlink()
                ret = super().write_to_fs(scheme, path, stream, is_executable)

            if scheme in ("platlib",
                          "purelib") and self.symlink_to is not None:
                path_dir = PurePath(path).parent
                to_top_dir = len(path_dir.parts) * ("..",)
                symlink_target = Path(*to_top_dir) / self.symlink_to / path
                orig_path = self.destdir_purelib / path
                full_target = orig_path.parent / symlink_target
                try:
                    # if we have a symlink already, we're probably doing
                    # an implementation chain, so follow it to the beginning
                    while full_target.is_symlink():
                        existing_symlink_target = full_target.readlink()
                        # we deliberately use logical normalization here
                        symlink_target = Path(
                            os.path.normpath(symlink_target.parent /
                                             existing_symlink_target))
                        full_target = orig_path.parent / symlink_target
                    if not filecmp.cmp(orig_path, full_target):
                        return ret
                except FileNotFoundError:
                    self.check_symlink_to()
                else:
                    orig_path.unlink()
                    orig_path.symlink_to(symlink_target)

            return ret

        def _compile_bytecode(self,
                              scheme: Scheme,
                              record: RecordEntry,
                              ) -> None:
            super()._compile_bytecode(scheme, record)

            if not self.symlink_pyc:
                return
            if scheme not in ("purelib", "platlib"):
                return
            if not record.path.endswith(".py"):
                return

            relative_path = PurePath(self.scheme_dict[scheme]) / record.path
            py = Path(self.destdir).joinpath(
                relative_path.relative_to(relative_path.anchor))
            pycs = [
                Path(importlib.util.cache_from_source(
                    py, optimization=opt if opt >= 1 else ""))
                for opt in self.bytecode_optimization_levels
            ]
            for pyc1, pyc2 in zip(pycs, pycs[1:]):
                if filecmp.cmp(pyc1, pyc2):
                    pyc2.unlink()
                    pyc2.symlink_to(pyc1.name)

    with WheelFile.open(wheel) as source:
        dest = DeduplicatingDestination(
            install_scheme_dict(args.prefix or DEFAULT_PREFIX,
                                source.distribution),
            str(args.interpreter),
            get_launcher_kind(),
            bytecode_optimization_levels=args.optimize,
            destdir=str(args.destdir),
            overwrite=args.overwrite,
            symlink_pyc=args.symlink_pyc,
            symlink_to=args.symlink_to,
        )
        logger.info(f"Installing {wheel} into {args.destdir}")
        install(source, dest, {})
        logger.info("Installation complete")



# Copyright 2021-2024 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import os
import stat

import portage


def installed_dynlibs(directory):
    """
    This traverses installed *.so symlinks to check if they point to
    regular files. If a symlink target is outside of the top directory,
    traversal follows the corresponding file inside the top directory
    if it exists, and otherwise stops following the symlink.
    """
    directory_prefix = f"{directory.rstrip(os.sep)}{os.sep}"
    for parent, _dirnames, filenames in os.walk(directory):
        for filename in filenames:
            if filename.endswith(".so"):
                filename_abs = os.path.join(parent, filename)
                target = filename_abs
                levels = 0
                while True:
                    try:
                        st = os.lstat(target)
                    except OSError:
                        break
                    if stat.S_ISREG(st.st_mode):
                        return True
                    elif stat.S_ISLNK(st.st_mode):
                        levels += 1
                        if levels == 40:
                            portage.writemsg(
                                f"too many levels of symbolic links: {filename_abs}\n",
                                noiselevel=-1,
                            )
                            break
                        target = portage.abssymlink(target)
                        if not target.startswith(directory_prefix):
                            # If target is outside the top directory, then follow the
                            # corresponding file inside the top directory if it exists,
                            # and otherwise stop following.
                            target = os.path.join(
                                directory_prefix, target.lstrip(os.sep)
                            )
                    else:
                        break
    return False


def check_dyn_libs_inconsistent(directory, provides):
    """Checks directory for whether any dynamic libraries were installed and
    if PROVIDES corresponds."""

    # Let's check if we've got inconsistent results.
    # If we're installing dynamic libraries (.so files), we should
    # really have a PROVIDES.
    # (This is a complementary check at the point of ingestion for the
    # creation check in doebuild.py)
    # Note: we could check a non-empty PROVIDES against the list of .sos,
    # but this doesn't gain us anything. We're interested in failure
    # to properly parse the installed files at all, which should really
    # be a global problem (e.g. bug #811462)
    return not provides and installed_dynlibs(directory)



# (c) 2022-2024 Michał Górny
# SPDX-License-Identifier: GPL-2.0-or-later

import importlib.util
import os
import os.path
import struct
import typing

from pathlib import Path


class PEP552Header:
    def __init__(self, pyc_f):
        # per PEP 552, the header consists of:
        # 4-byte magic (16-bit uint + b"\r\n")
        # 4-byte flags
        # depending on flags either:
        # - 4-byte timestamp + 4-byte size
        # - 8-byte hash

        self.header = pyc_f.read(16)
        if len(self.header) < 16:
            raise ValueError("Header too short")
        if self.header[:4] != importlib.util.MAGIC_NUMBER:
            raise ValueError("Invalid magic")

        # flags is a bitfield:
        # 0x01 - 0 for timestamp invalidation, 1 for hash invalidation
        # 0x02 - checked_source flag (for hash invalidation)
        # other bits are unused and must be zero
        self.flags, = struct.unpack_from("<L", self.header, 4)
        if self.flags & ~0x3:
            raise ValueError("Unexpected bits in flags set")
        self.invalidate_hash = bool(self.flags & 0x01)
        self.checked_source = bool(self.flags & 0x02)

        if self.checked_source and not self.invalidate_hash:
            raise ValueError(
                "checked_source flag set for timestamp invalidation")
        if self.invalidate_hash:
            # hash-based invalidation
            # (Python is passing this hash as a bytestring)
            self.py_hash = self.header[8:16]
            assert len(self.py_hash) == 8
        else:
            # timestamp-based invalidation
            self.py_timestamp, self.py_size = (
                struct.unpack_from("<LL", self.header, 8))


def qa_verify_pyc(destdir: Path, sitedirs: typing.Iterable[Path]):
    missing_pyc = []
    invalid_pyc = []
    mismatched_pyc = []
    stray_pyc = []

    for sitedir in sitedirs:
        top_path = destdir / sitedir.relative_to(sitedir.root)
        if not top_path.is_dir():
            continue

        py_files: typing.Set[Path] = set()
        pyc_files: typing.Set[Path] = set()

        for path, dirs, files in os.walk(top_path):
            for fn in files:
                if fn.endswith(".py"):
                    py_files.add(Path(path) / fn)
                elif fn.endswith((".pyc", ".pyo")):
                    pyc_files.add(Path(path) / fn)

        for py in py_files:
            py_stat = py.stat()

            for opt in ("", 1, 2):
                pyc = Path(
                    importlib.util.cache_from_source(py, optimization=opt))
                # 1. check for missing .pyc files
                if pyc not in pyc_files:
                    missing_pyc.append((pyc, py))
                    continue

                pyc_files.remove(pyc)
                # 2. check the header
                with pyc.open("rb") as f:
                    try:
                        header = PEP552Header(f)
                    except ValueError:
                        invalid_pyc.append((pyc, py))
                        continue

                # 3. check whether .pyc matches the .py file
                if header.invalidate_hash:
                    # NB: even though !checked_source implies that Python
                    # does not verify the .pyc validity, we do because
                    # per that mode we're actually responsible for ensuring
                    # that the file is valid
                    with py.open("rb") as f:
                        py_hash = importlib.util.source_hash(f.read())
                    if py_hash != header.py_hash:
                        mismatched_pyc.append((pyc, py, "hash"))
                else:
                    if int(py_stat.st_mtime) != header.py_timestamp:
                        mismatched_pyc.append((pyc, py, "timestamp"))
                    if py_stat.st_size != header.py_size:
                        mismatched_pyc.append((pyc, py, "size"))

        # 4. any remaining .pyc files are stray
        stray_pyc.extend((pyc,) for pyc in pyc_files)

    return {
        "missing": missing_pyc,
        "invalid": invalid_pyc,
        "mismatched": mismatched_pyc,
        "stray": stray_pyc,
    }



# SPDX-License-Identifier: Apache-2.0
# Copyright 2019 The Meson development team
# Copyright © 2023-2025 Intel Corporation

from __future__ import annotations

import typing as T

from mesonbuild.templates.sampleimpl import FileImpl

if T.TYPE_CHECKING:
    from ..minit import Arguments


lib_rust_template = '''#![crate_name = "{crate_file}"]

/* This function will not be exported and is not
 * directly callable by users of this library.
 */
fn internal_function() -> i32 {{
    return 0;
}}

pub fn {function_name}() -> i32 {{
    return internal_function();
}}

#[cfg(test)]
mod tests {{
    use super::*;

    #[test]
    fn test_function() {{
        assert_eq!({function_name}(), 0);
    }}
}}
'''


lib_rust_meson_template = '''project(
  '{project_name}',
  'rust',
  version : '{version}',
  meson_version : '>= {meson_version}',
  default_options : ['rust_std=2021', 'warning_level=3'],
)

rust = import('rust')

dependencies = [{dependencies}
]

lib = static_library(
  '{lib_name}',
  '{source_file}',
  dependencies : dependencies,
  install : true,
)

rust.test('{test_name}', lib)

# Make this library usable as a Meson subproject.
{ltoken}_dep = declare_dependency(
  include_directories : include_directories('.'),
  dependencies : dependencies,
  link_with : lib,
)
meson.override_dependency('{project_name}', {ltoken}_dep)
'''

hello_rust_template = '''
fn main() {{
    let project_name = "{project_name}";
    println!("This is project {{}}.\\n", project_name);
}}
'''

hello_rust_meson_template = '''project(
  '{project_name}',
  'rust',
  version : '{version}',
  meson_version : '>= {meson_version}',
  default_options : ['rust_std=2021', 'warning_level=3'],
)

dependencies = [{dependencies}
]

exe = executable(
  '{exe_name}',
  '{source_name}',
  dependencies : dependencies,
  install : true,
)

test('basic', exe)
'''


class RustProject(FileImpl):

    source_ext = 'rs'
    exe_template = hello_rust_template
    exe_meson_template = hello_rust_meson_template
    lib_template = lib_rust_template
    lib_test_template = None
    lib_meson_template = lib_rust_meson_template

    def __init__(self, args: Arguments):
        super().__init__(args)
        self.meson_version = '1.3.0'

    def lib_kwargs(self) -> T.Dict[str, str]:
        kwargs = super().lib_kwargs()
        kwargs['crate_file'] = self.lowercase_token
        return kwargs



# Copyright 2005-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2
# Author(s): Brian Harring (ferringb@gentoo.org)

import os as _os
from portage.cache import template
from portage import os


class FsBased(template.database):
    """template wrapping fs needed options, and providing _ensure_access as a way to
    attempt to ensure files have the specified owners/perms"""

    def __init__(self, *args, **config):
        for x, y in (("gid", -1), ("perms", 0o644)):
            if x in config:
                # Since Python 3.4, chown requires int type (no proxies).
                setattr(self, "_" + x, int(config[x]))
                del config[x]
            else:
                setattr(self, "_" + x, y)
        super().__init__(*args, **config)

        if self.label.startswith(os.path.sep):
            # normpath.
            self.label = os.path.sep + os.path.normpath(self.label).lstrip(os.path.sep)

    def _ensure_access(self, path, mtime=-1):
        """returns true or false if it's able to ensure that path is properly chmod'd and chowned.
        if mtime is specified, attempts to ensure that's correct also"""
        from portage.exception import PortageException
        from portage.util import apply_permissions

        try:
            apply_permissions(path, gid=self._gid, mode=self._perms)
            if mtime != -1:
                mtime = int(mtime)
                os.utime(path, (mtime, mtime))
        except (PortageException, OSError):
            return False
        return True

    def _ensure_dirs(self, path=None):
        """with path!=None, ensure beyond self.location.  otherwise, ensure self.location"""
        from portage.util import ensure_dirs, apply_permissions

        if path:
            path = os.path.dirname(path)
            base = self.location
        else:
            path = self.location
            base = "/"

        for d in path.lstrip(os.path.sep).rstrip(os.path.sep).split(os.path.sep):
            base = os.path.join(base, d)
            if ensure_dirs(base):
                # We only call apply_permissions if ensure_dirs created
                # a new directory, so as not to interfere with
                # permissions of existing directories.
                mode = self._perms
                if mode == -1:
                    mode = 0
                mode |= 0o755
                apply_permissions(base, mode=mode, gid=self._gid)

    def _prune_empty_dirs(self):
        all_dirs = []
        for parent, dirs, files in os.walk(self.location):
            for x in dirs:
                all_dirs.append(_os.path.join(parent, x))
        while all_dirs:
            try:
                _os.rmdir(all_dirs.pop())
            except OSError:
                pass


def gen_label(base, label):
    """if supplied label is a path, generate a unique label based upon label, and supplied base path"""
    if label.find(os.path.sep) == -1:
        return label
    label = label.strip('"').strip("'")
    label = os.path.join(*(label.rstrip(os.path.sep).split(os.path.sep)))
    tail = os.path.split(label)[1]
    return f"{tail}-{abs(label.__hash__()):X}"



# test_varExpand.py -- Portage Unit Testing Functionality
# Copyright 2006-2017 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.util import varexpand


class VarExpandTestCase(TestCase):
    def testVarExpandPass(self):
        varDict = {"a": "5", "b": "7", "c": "-5"}
        for key in varDict:
            result = varexpand(f"${key}", varDict)

            self.assertFalse(
                result != varDict[key],
                msg="Got %s != %s, from varexpand(%s, %s)"
                % (result, varDict[key], "$%s" % key, varDict),
            )
            result = varexpand("${%s}" % key, varDict)
            self.assertFalse(
                result != varDict[key],
                msg="Got %s != %s, from varexpand(%s, %s)"
                % (result, varDict[key], "${%s}" % key, varDict),
            )

    def testVarExpandBackslashes(self):
        r"""
        We want to behave like bash does when expanding a variable
        assignment in a sourced file, in which case it performs
        backslash removal for \\ and \$ but nothing more. It also
        removes escaped newline characters. Note that we don't
        handle escaped quotes here, since getconfig() uses shlex
        to handle that earlier.
        """

        varDict = {}
        tests = [
            ("\\", "\\"),
            ("\\\\", "\\"),
            ("\\\\\\", "\\\\"),
            ("\\\\\\\\", "\\\\"),
            ("\\$", "$"),
            ("\\\\$", "\\$"),
            ("\\a", "\\a"),
            ("\\b", "\\b"),
            ("\\n", "\\n"),
            ("\\r", "\\r"),
            ("\\t", "\\t"),
            ("\\\n", ""),
            ('\\"', '\\"'),
            ("\\'", "\\'"),
        ]
        for test in tests:
            result = varexpand(test[0], varDict)
            self.assertFalse(
                result != test[1],
                msg=f"Got {result} != {test[1]} from varexpand({test[0]}, {varDict})",
            )

    def testVarExpandDoubleQuotes(self):
        varDict = {"a": "5"}
        tests = [('"${a}"', '"5"')]
        for test in tests:
            result = varexpand(test[0], varDict)
            self.assertFalse(
                result != test[1],
                msg=f"Got {result} != {test[1]} from varexpand({test[0]}, {varDict})",
            )

    def testVarExpandSingleQuotes(self):
        varDict = {"a": "5"}
        tests = [("'${a}'", "'${a}'")]
        for test in tests:
            result = varexpand(test[0], varDict)
            self.assertFalse(
                result != test[1],
                msg=f"Got {result} != {test[1]} from varexpand({test[0]}, {varDict})",
            )

    def testVarExpandFail(self):
        varDict = {"a": "5", "b": "7", "c": "15"}

        testVars = ["fail"]

        for var in testVars:
            result = varexpand(f"${var}", varDict)
            self.assertFalse(
                len(result),
                msg="Got %s == %s, from varexpand(%s, %s)"
                % (result, var, "$%s" % var, varDict),
            )

            result = varexpand("${%s}" % var, varDict)
            self.assertFalse(
                len(result),
                msg="Got %s == %s, from varexpand(%s, %s)"
                % (result, var, "${%s}" % var, varDict),
            )



import ast
from contextlib import contextmanager
import hashlib
import logging
import os
import sys

from pathlib import Path
import re

log = logging.getLogger(__name__)

from .versionno import normalise_version

class Module(object):
    """This represents the module/package that we are going to distribute
    """
    in_namespace_package = False
    namespace_package_name = None

    def __init__(self, name, directory=Path()):
        self.name = name

        # It must exist either as a .py file or a directory, but not both
        name_as_path = name.replace('.', os.sep)
        pkg_dir = directory / name_as_path
        py_file = directory / (name_as_path+'.py')
        src_pkg_dir = directory / 'src' / name_as_path
        src_py_file = directory / 'src' / (name_as_path+'.py')

        existing = set()
        if pkg_dir.is_dir():
            self.path = pkg_dir
            self.is_package = True
            self.prefix = ''
            existing.add(pkg_dir)
        if py_file.is_file():
            self.path = py_file
            self.is_package = False
            self.prefix = ''
            existing.add(py_file)
        if src_pkg_dir.is_dir():
            self.path = src_pkg_dir
            self.is_package = True
            self.prefix = 'src'
            existing.add(src_pkg_dir)
        if src_py_file.is_file():
            self.path = src_py_file
            self.is_package = False
            self.prefix = 'src'
            existing.add(src_py_file)

        if len(existing) > 1:
            raise ValueError(
                "Multiple files or folders could be module {}: {}"
                .format(name, ", ".join([str(p) for p in sorted(existing)]))
            )
        elif not existing:
            raise ValueError("No file/folder found for module {}".format(name))

        self.source_dir = directory / self.prefix

        if '.' in name:
            self.namespace_package_name = name.rpartition('.')[0]
            self.in_namespace_package = True

    @property
    def file(self):
        if self.is_package:
            return self.path / '__init__.py'
        else:
            return self.path

    @property
    def version_files(self):
        """Files which will be parsed to find a version number

        Files later in this list take precedence over earlier ones.
        """
        if self.is_package:
            paths = [self.path / '__init__.py']
            for filename in ('version.py', '_version.py', '__version__.py'):
                if (self.path / filename).is_file():
                    paths.insert(0, self.path / filename)
            return paths
        else:
            return [self.path]

    def iter_files(self):
        """Iterate over the files contained in this module.

        Yields absolute paths - caller may want to make them relative.
        Excludes any __pycache__ and *.pyc files.
        """
        def _include(path):
            name = os.path.basename(path)
            if (name == '__pycache__') or name.endswith('.pyc'):
                return False
            return True

        if self.is_package:
            # Ensure we sort all files and directories so the order is stable
            for dirpath, dirs, files in os.walk(str(self.path)):
                for file in sorted(files):
                    full_path = os.path.join(dirpath, file)
                    if _include(full_path):
                        yield full_path

                dirs[:] = [d for d in sorted(dirs) if _include(d)]

        else:
            yield str(self.path)

class ProblemInModule(ValueError): pass
class NoDocstringError(ProblemInModule): pass
class NoVersionError(ProblemInModule): pass
class InvalidVersion(ProblemInModule): pass

class VCSError(Exception):
    def __init__(self, msg, directory):
        self.msg = msg
        self.directory = directory

    def __str__(self):
        return self.msg + ' ({})'.format(self.directory)


@contextmanager
def _module_load_ctx():
    """Preserve some global state that modules might change at import time.

    - Handlers on the root logger.
    """
    logging_handlers = logging.root.handlers[:]
    try:
        yield
    finally:
        logging.root.handlers = logging_handlers

def get_docstring_and_version_via_ast(target):
    """
    Return a tuple like (docstring, version) for the given module,
    extracted by parsing its AST.
    """
    version = None
    for target_path in target.version_files:
        # read as bytes to enable custom encodings
        with target_path.open('rb') as f:
            node = ast.parse(f.read())
        for child in node.body:
            if is_version_str_assignment(child):
                if sys.version_info >= (3, 8):
                    version = child.value.value
                else:
                    version = child.value.s
                break
    return ast.get_docstring(node), version


def is_version_str_assignment(node):
    """Check if *node* is a simple string assignment to __version__"""
    if not isinstance(node, (ast.Assign, ast.AnnAssign)):
        return False
    constant_type = ast.Constant if sys.version_info >= (3, 8) else ast.Str
    if not isinstance(node.value, constant_type):
        return False
    targets = (node.target,) if isinstance(node, ast.AnnAssign) else node.targets
    for target in targets:
        if isinstance(target, ast.Name) and target.id == "__version__":
            return True
    return False


# To ensure we're actually loading the specified file, give it a unique name to
# avoid any cached import. In normal use we'll only load one module per process,
# so it should only matter for the tests, but we'll do it anyway.
_import_i = 0


def get_docstring_and_version_via_import(target):
    """
    Return a tuple like (docstring, version) for the given module,
    extracted by importing the module and pulling __doc__ & __version__
    from it.
    """
    global _import_i
    _import_i += 1

    log.debug("Loading module %s", target.file)
    from importlib.util import spec_from_file_location, module_from_spec
    mod_name = 'flit_core.dummy.import%d' % _import_i
    spec = spec_from_file_location(mod_name, target.file)
    with _module_load_ctx():
        m = module_from_spec(spec)
        # Add the module to sys.modules to allow relative imports to work.
        # importlib has more code around this to handle the case where two
        # threads are trying to load the same module at the same time, but Flit
        # should always be running a single thread, so we won't duplicate that.
        sys.modules[mod_name] = m
        try:
            spec.loader.exec_module(m)
        finally:
            sys.modules.pop(mod_name, None)

    docstring = m.__dict__.get('__doc__', None)
    version = m.__dict__.get('__version__', None)
    return docstring, version


def get_info_from_module(target, for_fields=('version', 'description')):
    """Load the module/package, get its docstring and __version__
    """
    if not for_fields:
        return {}

    # What core metadata calls Summary, PEP 621 calls description
    want_summary = 'description' in for_fields
    want_version = 'version' in for_fields

    log.debug("Loading module %s", target.file)

    # Attempt to extract our docstring & version by parsing our target's
    # AST, falling back to an import if that fails. This allows us to
    # build without necessarily requiring that our built package's
    # requirements are installed.
    docstring, version = get_docstring_and_version_via_ast(target)
    if (want_summary and not docstring) or (want_version and not version):
        docstring, version = get_docstring_and_version_via_import(target)

    res = {}

    if want_summary:
        if (not docstring) or not docstring.strip():
            raise NoDocstringError(
                'Flit cannot package module without docstring, or empty docstring. '
                'Please add a docstring to your module ({}).'.format(target.file)
            )
        res['summary'] = docstring.lstrip().splitlines()[0]

    if want_version:
        res['version'] = check_version(version)

    return res

def check_version(version):
    """
    Check whether a given version string match PEP 440, and do normalisation.

    Raise InvalidVersion/NoVersionError with relevant information if
    version is invalid.

    Log a warning if the version is not canonical with respect to PEP 440.

    Returns the version in canonical PEP 440 format.
    """
    if not version:
        raise NoVersionError('Cannot package module without a version string. '
                             'Please define a `__version__ = "x.y.z"` in your module.')
    if not isinstance(version, str):
        raise InvalidVersion('__version__ must be a string, not {}.'
                                .format(type(version)))

    # Import here to avoid circular import
    version = normalise_version(version)

    return version


script_template = """\
#!{interpreter}
# -*- coding: utf-8 -*-
import re
import sys
from {module} import {import_name}
if __name__ == '__main__':
    sys.argv[0] = re.sub(r'(-script\\.pyw|\\.exe)?$', '', sys.argv[0])
    sys.exit({func}())
"""

def parse_entry_point(ep):
    """Check and parse a 'package.module:func' style entry point specification.

    Returns (modulename, funcname)
    """
    if ':' not in ep:
        raise ValueError("Invalid entry point (no ':'): %r" % ep)
    mod, func = ep.split(':')

    for piece in func.split('.'):
        if not piece.isidentifier():
            raise ValueError("Invalid entry point: %r is not an identifier" % piece)
    for piece in mod.split('.'):
        if not piece.isidentifier():
            raise ValueError("Invalid entry point: %r is not a module path" % piece)

    return mod, func

def write_entry_points(d, fp):
    """Write entry_points.txt from a two-level dict

    Sorts on keys to ensure results are reproducible.
    """
    for group_name in sorted(d):
        fp.write(u'[{}]\n'.format(group_name))
        group = d[group_name]
        for name in sorted(group):
            val = group[name]
            fp.write(u'{}={}\n'.format(name, val))
        fp.write(u'\n')

def hash_file(path, algorithm='sha256'):
    with open(path, 'rb') as f:
        h = hashlib.new(algorithm, f.read())
    return h.hexdigest()

def normalize_file_permissions(st_mode):
    """Normalize the permission bits in the st_mode field from stat to 644/755

    Popular VCSs only track whether a file is executable or not. The exact
    permissions can vary on systems with different umasks. Normalising
    to 644 (non executable) or 755 (executable) makes builds more reproducible.
    """
    # Set 644 permissions, leaving higher bits of st_mode unchanged
    new_mode = (st_mode | 0o644) & ~0o133
    if st_mode & 0o100:
        new_mode |= 0o111  # Executable: 644 -> 755
    return new_mode

class Metadata(object):

    summary = None
    home_page = None
    author = None
    author_email = None
    maintainer = None
    maintainer_email = None
    license = None
    license_expression = None
    description = None
    keywords = None
    download_url = None
    requires_python = None
    description_content_type = None

    platform = ()
    supported_platform = ()
    classifiers = ()
    provides = ()
    requires = ()
    obsoletes = ()
    project_urls = ()
    provides_dist = ()
    requires_dist = ()
    obsoletes_dist = ()
    requires_external = ()
    provides_extra = ()
    license_files = ()
    dynamic = ()

    metadata_version = "2.4"

    def __init__(self, data):
        data = data.copy()
        self.name = data.pop('name')
        self.version = data.pop('version')

        for k, v in data.items():
            assert hasattr(self, k), "data does not have attribute '{}'".format(k)
            setattr(self, k, v)

    def _normalise_field_name(self, n):
        return n.lower().replace('-', '_')

    def _extract_extras(self, req):
        match = re.search(r'\[([^]]*)\]', req)
        if match:
            list_str = match.group(1)
            return [item.strip() for item in list_str.split(',')]
        else:
            return None

    def _normalise_requires_dist(self, req):
        extras = self._extract_extras(req)
        if extras:
            normalised_extras = [normalise_core_metadata_name(extra) for extra in extras]
            normalised_extras_str = ', '.join(normalised_extras)
            normalised_req = re.sub(r'\[([^]]*)\]', f"[{normalised_extras_str}]", req)
            return normalised_req
        else:
            return req

    def write_metadata_file(self, fp):
        """Write out metadata in the email headers format"""
        fields = [
            'Metadata-Version',
            'Name',
            'Version',
        ]
        optional_fields = [
            'Summary',
            'Home-page',
            'Keywords',
            'Author',
            'Author-email',
            'Maintainer',
            'Maintainer-email',
            'Requires-Python',
            'Description-Content-Type',
        ]

        for field in fields:
            value = getattr(self, self._normalise_field_name(field))
            fp.write(u"{}: {}\n".format(field, value))

        for field in optional_fields:
            value = getattr(self, self._normalise_field_name(field))
            if value is not None:
                # TODO: verify which fields can be multiline
                # The spec has multiline examples for Author, Maintainer &
                # License (& Description, but we put that in the body)
                # Indent following lines with 8 spaces:
                value = '\n        '.join(value.splitlines())
                fp.write(u"{}: {}\n".format(field, value))


        license_expr = getattr(self, self._normalise_field_name("License-Expression"))
        license = getattr(self, self._normalise_field_name("License"))
        if license_expr:
            fp.write(u'License-Expression: {}\n'.format(license_expr))
        elif license:  # Deprecated, superseded by License-Expression
            fp.write(u'License: {}\n'.format(license))

        for clsfr in self.classifiers:
            fp.write(u'Classifier: {}\n'.format(clsfr))

        for file in self.license_files:
            fp.write(u'License-File: {}\n'.format(file))

        for req in self.requires_dist:
            normalised_req = self._normalise_requires_dist(req)
            fp.write(u'Requires-Dist: {}\n'.format(normalised_req))

        for url in self.project_urls:
            fp.write(u'Project-URL: {}\n'.format(url))

        for extra in self.provides_extra:
            normalised_extra = normalise_core_metadata_name(extra)
            fp.write(u'Provides-Extra: {}\n'.format(normalised_extra))

        if self.description is not None:
            fp.write(u'\n' + self.description + u'\n')

    @property
    def supports_py2(self):
        """Return True if Requires-Python indicates Python 2 support."""
        for part in (self.requires_python or "").split(","):
            if re.search(r"^\s*(>=?|~=|===?)?\s*[3-9]", part):
                return False
        return True


def make_metadata(module, ini_info):
    md_dict = {'name': module.name, 'provides': [module.name]}
    md_dict.update(get_info_from_module(module, ini_info.dynamic_metadata))
    md_dict.update(ini_info.metadata)
    return Metadata(md_dict)


def normalise_core_metadata_name(name):
    """Normalise a project or extra name (as in PEP 503, also PEP 685)"""
    return re.sub(r"[-_.]+", "-", name).lower()


def normalize_dist_name(name: str, version: str) -> str:
    """Normalizes a name and a PEP 440 version

    The resulting string is valid as dist-info folder name
    and as first part of a wheel filename

    See https://packaging.python.org/specifications/binary-distribution-format/#escaping-and-unicode
    """
    normalized_name = re.sub(r'[-_.]+', '_', name, flags=re.UNICODE).lower()
    assert check_version(version) == version
    assert '-' not in version, 'Normalized versions can’t have dashes'
    return '{}-{}'.format(normalized_name, version)


def dist_info_name(distribution, version):
    """Get the correct name of the .dist-info folder"""
    return normalize_dist_name(distribution, version) + '.dist-info'


def walk_data_dir(data_directory):
    """Iterate over the files in the given data directory.

    Yields paths prefixed with data_directory - caller may want to make them
    relative to that. Excludes any __pycache__ subdirectories.
    """
    if data_directory is None:
        return

    for dirpath, dirs, files in os.walk(data_directory):
        for file in sorted(files):
            full_path = os.path.join(dirpath, file)
            yield full_path

        dirs[:] = [d for d in sorted(dirs) if d != '__pycache__']



# Copyright 1998-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import logging
import operator
import sys
import signal

from _emerge.ProgressHandler import ProgressHandler

import portage
from portage import eapi_is_supported
from portage import os
from portage.cache.cache_errors import CacheError
from portage.eclass_cache import hashed_path
from portage.util import writemsg_level


def action_metadata(settings, portdb, myopts, porttrees=None):
    if porttrees is None:
        porttrees = portdb.porttrees
    portage.writemsg_stdout("\n>>> Updating Portage cache\n")
    cachedir = os.path.normpath(settings.depcachedir)
    if cachedir in [
        "/",
        "/bin",
        "/dev",
        "/etc",
        "/home",
        "/lib",
        "/opt",
        "/proc",
        "/root",
        "/sbin",
        "/sys",
        "/tmp",
        "/usr",
        "/var",
    ]:
        print(
            (
                "!!! PORTAGE_DEPCACHEDIR IS SET TO A PRIMARY "
                "ROOT DIRECTORY ON YOUR SYSTEM.\n"
                f"!!! This is ALMOST CERTAINLY NOT what you want: '{cachedir}'",
            ),
            file=sys.stderr,
        )
        sys.exit(73)
    if not os.path.exists(cachedir):
        os.makedirs(cachedir)

    auxdbkeys = portdb._known_keys

    class TreeData:
        __slots__ = ("dest_db", "eclass_db", "path", "src_db", "valid_nodes")

        def __init__(self, dest_db, eclass_db, path, src_db):
            self.dest_db = dest_db
            self.eclass_db = eclass_db
            self.path = path
            self.src_db = src_db
            self.valid_nodes = set()

    porttrees_data = []
    for path in porttrees:
        src_db = portdb._pregen_auxdb.get(path)
        if src_db is None:
            # portdbapi does not populate _pregen_auxdb
            # when FEATURES=metadata-transfer is enabled
            src_db = portdb._create_pregen_cache(path)

        if src_db is not None:
            eclass_db = portdb.repositories.get_repo_for_location(path).eclass_db
            # Update eclass data which may be stale after sync.
            eclass_db.update_eclasses()
            porttrees_data.append(TreeData(portdb.auxdb[path], eclass_db, path, src_db))

    quiet = (
        settings.get("TERM") == "dumb" or "--quiet" in myopts or not sys.stdout.isatty()
    )

    onProgress = None
    if not quiet:
        progressBar = portage.output.TermProgressBar()
        progressHandler = ProgressHandler()
        onProgress = progressHandler.onProgress

        def display():
            progressBar.set(progressHandler.curval, progressHandler.maxval)

        progressHandler.display = display

        def sigwinch_handler(signum, frame):
            lines, progressBar.term_columns = portage.output.get_term_size()

        signal.signal(signal.SIGWINCH, sigwinch_handler)

    # Temporarily override portdb.porttrees so portdb.cp_all()
    # will only return the relevant subset.
    portdb_porttrees = portdb.porttrees
    portdb.porttrees = (tree_data.path for tree_data in porttrees_data)
    try:
        cp_all = portdb.cp_all()
    finally:
        portdb.porttrees = portdb_porttrees

    curval = 0
    maxval = len(cp_all)
    if onProgress is not None:
        onProgress(maxval, curval)

    # TODO: Display error messages, but do not interfere with the progress bar.
    # Here's how:
    #  1) erase the progress bar
    #  2) show the error message
    #  3) redraw the progress bar on a new line

    for cp in cp_all:
        for tree_data in porttrees_data:
            src_chf = tree_data.src_db.validation_chf
            dest_chf = tree_data.dest_db.validation_chf
            dest_chf_key = f"_{dest_chf}_"
            dest_chf_getter = operator.attrgetter(dest_chf)

            for cpv in portdb.cp_list(cp, mytree=tree_data.path):
                tree_data.valid_nodes.add(cpv)
                try:
                    src = tree_data.src_db[cpv]
                except (CacheError, KeyError):
                    continue

                ebuild_location = portdb.findname(cpv, mytree=tree_data.path)
                if ebuild_location is None:
                    continue
                ebuild_hash = hashed_path(ebuild_location)

                try:
                    if not tree_data.src_db.validate_entry(
                        src, ebuild_hash, tree_data.eclass_db
                    ):
                        continue
                except CacheError:
                    continue

                eapi = src.get("EAPI")
                if not eapi:
                    eapi = "0"
                eapi_supported = eapi_is_supported(eapi)
                if not eapi_supported:
                    continue

                dest = None
                try:
                    dest = tree_data.dest_db[cpv]
                except (KeyError, CacheError):
                    pass

                for d in (src, dest):
                    if d is not None and d.get("EAPI") in ("", "0"):
                        del d["EAPI"]

                if src_chf != "mtime":
                    # src may contain an irrelevant _mtime_ which corresponds
                    # to the time that the cache entry was written
                    src.pop("_mtime_", None)

                if src_chf != dest_chf:
                    # populate src entry with dest_chf_key
                    # (the validity of the dest_chf that we generate from the
                    # ebuild here relies on the fact that we already used
                    # validate_entry to validate the ebuild with src_chf)
                    src[dest_chf_key] = dest_chf_getter(ebuild_hash)

                if dest is not None:
                    if not (
                        dest.get(dest_chf_key) == src[dest_chf_key]
                        and tree_data.eclass_db.validate_and_rewrite_cache(
                            dest["_eclasses_"],
                            tree_data.dest_db.validation_chf,
                            tree_data.dest_db.store_eclass_paths,
                        )
                        is not None
                        and set(dest["_eclasses_"]) == set(src["_eclasses_"])
                    ):
                        dest = None
                    else:
                        # We don't want to skip the write unless we're really
                        # sure that the existing cache is identical, so don't
                        # trust _mtime_ and _eclasses_ alone.
                        cache_is_identical = (
                            True for k in auxdbkeys if dest.get(k, "") != src.get(k, "")
                        )
                        if any(cache_is_identical):
                            dest = None
                if dest is not None:
                    # The existing data is valid and identical,
                    # so there's no need to overwrite it.
                    continue

                try:
                    tree_data.dest_db[cpv] = src
                except CacheError:
                    # ignore it; can't do anything about it.
                    pass

        curval += 1
        if onProgress is not None:
            onProgress(maxval, curval)

    if onProgress is not None:
        onProgress(maxval, curval)

    for tree_data in porttrees_data:
        try:
            dead_nodes = set(tree_data.dest_db)
        except CacheError as e:
            writemsg_level(
                (
                    "Error listing cache entries for "
                    f"'{tree_data.path}': {e}, continuing...\n"
                ),
                level=logging.ERROR,
                noiselevel=-1,
            )
            del e
        else:
            dead_nodes.difference_update(tree_data.valid_nodes)
            for cpv in dead_nodes:
                try:
                    del tree_data.dest_db[cpv]
                except (KeyError, CacheError):
                    pass

    if not quiet:
        # make sure the final progress is displayed
        progressHandler.display()
        print()
        signal.signal(signal.SIGWINCH, signal.SIG_DFL)

    portdb.flush_cache()
    sys.stdout.flush()



# Copyright 1999-2013 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import gc


def clear_caches(trees):
    for d in trees.values():
        d["porttree"].dbapi.melt()
        d["porttree"].dbapi._aux_cache.clear()
        d["bintree"].dbapi._clear_cache()
        if d["vartree"].dbapi._linkmap is None:
            # preserve-libs is entirely disabled
            pass
        else:
            d["vartree"].dbapi._linkmap._clear_cache()
    gc.collect()



# Copyright 2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2


from portage import (
    _encodings,
    _unicode_encode,
    os,
)
from portage.dep.soname.parse import parse_soname_deps
from portage.util._dyn_libs.NeededEntry import NeededEntry


def _get_all_provides(vardb):
    """
    Get all of the sonames provided by all of the installed packages.
    This does not bother to acquire a lock, since its pretty safe to
    assume that any packages merged or unmerged while this function
    is running must be irrelevant.

    @param vardb: an installed package database
    @type vardb: vardbapi
    @rtype: frozenset
    @return: a frozenset od SonameAtom instances provided by all
            installed packages
    """

    all_provides = []

    for cpv in vardb.cpv_all():
        try:
            (provides,) = vardb.aux_get(cpv, ["PROVIDES"])
        except KeyError:
            # Since we don't hold a lock, assume this is due to a
            # concurrent unmerge, and PROVIDES from the unmerged package
            # are most likely negligible due to topologically sorted
            # merge order. Also, note that it's possible for aux_get
            # to succeed and return empty PROVIDES metadata if the file
            # disappears (due to unmerge) before it can be read.
            pass
        else:
            if provides:
                all_provides.extend(parse_soname_deps(provides))

    return frozenset(all_provides)


def _get_unresolved_soname_deps(metadata_dir, all_provides):
    """
    Get files with unresolved soname dependencies.

    @param metadata_dir: directory containing package metadata files
            named REQUIRES and NEEDED.ELF.2
    @type metadata_dir: str
    @param all_provides: a frozenset on SonameAtom instances provided by
            all installed packages
    @type all_provides: frozenset
    @rtype: list
    @return: list of tuple(filename, tuple(unresolved sonames))
    """
    try:
        with open(
            _unicode_encode(
                os.path.join(metadata_dir, "REQUIRES"),
                encoding=_encodings["fs"],
                errors="strict",
            ),
            encoding=_encodings["repo.content"],
            errors="strict",
        ) as f:
            requires = frozenset(parse_soname_deps(f.read()))
    except OSError:
        return []

    unresolved_by_category = {}
    for atom in requires:
        if atom not in all_provides:
            unresolved_by_category.setdefault(atom.multilib_category, set()).add(
                atom.soname
            )

    needed_filename = os.path.join(metadata_dir, "NEEDED.ELF.2")
    with open(
        _unicode_encode(needed_filename, encoding=_encodings["fs"], errors="strict"),
        encoding=_encodings["repo.content"],
        errors="strict",
    ) as f:
        needed = f.readlines()

    unresolved_by_file = []
    for l in needed:
        l = l.rstrip("\n")
        if not l:
            continue
        entry = NeededEntry.parse(needed_filename, l)
        missing = unresolved_by_category.get(entry.multilib_category)
        if not missing:
            continue
        # NOTE: This can contain some false positives in the case of
        # missing DT_RPATH settings, since it's possible that a subset
        # package files have the desired DT_RPATH settings. However,
        # since reported sonames are unresolved for at least some file(s),
        # false positives or this sort should not be not too annoying.
        missing = [soname for soname in entry.needed if soname in missing]
        if missing:
            unresolved_by_file.append((entry.filename, tuple(missing)))

    return unresolved_by_file



# Copyright 2015 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import struct


def decode_uint16_be(data):
    """
    Decode an unsigned 16-bit integer with big-endian encoding.

    @param data: string of bytes of length 2
    @type data: bytes
    @rtype: int
    @return: unsigned integer value of the decoded data
    """
    return struct.unpack_from(">H", data)[0]


def decode_uint16_le(data):
    """
    Decode an unsigned 16-bit integer with little-endian encoding.

    @param data: string of bytes of length 2
    @type data: bytes
    @rtype: int
    @return: unsigned integer value of the decoded data
    """
    return struct.unpack_from("<H", data)[0]


def decode_uint32_be(data):
    """
    Decode an unsigned 32-bit integer with big-endian encoding.

    @param data: string of bytes of length 4
    @type data: bytes
    @rtype: int
    @return: unsigned integer value of the decoded data
    """
    return struct.unpack_from(">I", data)[0]


def decode_uint32_le(data):
    """
    Decode an unsigned 32-bit integer with little-endian encoding.

    @param data: string of bytes of length 4
    @type data: bytes
    @rtype: int
    @return: unsigned integer value of the decoded data
    """
    return struct.unpack_from("<I", data)[0]



# Copyright 2013-2024 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import multiprocessing

import portage
from portage import os
from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import ResolverPlayground
from portage.package.ebuild._ipc.QueryCommand import QueryCommand
from portage.util._async.ForkProcess import ForkProcess
from portage.util._async.TaskScheduler import TaskScheduler
from _emerge.Package import Package
from _emerge.PipeReader import PipeReader


class DoebuildFdPipesTestCase(TestCase):
    output_fd = 200

    def testDoebuild(self):
        """
        Invoke portage.doebuild() with the fd_pipes parameter, and
        check that the expected output appears in the pipe. This
        functionality is not used by portage internally, but it is
        supported for API consumers (see bug #475812).
        """

        output_fd = self.output_fd
        ebuild_body = ["S=${WORKDIR}"]
        for phase_func, default in (
            ("pkg_info", False),
            ("pkg_nofetch", False),
            ("pkg_pretend", False),
            ("pkg_setup", False),
            ("pkg_config", False),
            ("src_unpack", False),
            ("src_prepare", True),
            ("src_configure", False),
            ("src_compile", False),
            ("src_test", False),
            ("src_install", False),
        ):
            ebuild_body.append(
                ("%s() { %secho ${EBUILD_PHASE}" " 1>&%s; }")
                % (phase_func, "default; " if default else "", output_fd)
            )

        ebuild_body.append("")
        ebuild_body = "\n".join(ebuild_body)

        ebuilds = {
            "app-misct/foo-1": {
                "EAPI": "8",
                "IUSE": "+foo +bar",
                "REQUIRED_USE": "|| ( foo bar )",
                "MISC_CONTENT": ebuild_body,
            }
        }

        # Populate configdict["pkg"]["USE"] with something arbitrary in order
        # to try and trigger bug 675748 in doebuild _validate_deps.
        arbitrary_package_use = "baz"

        user_config = {
            # In order to trigger bug 675748, package.env must be non-empty,
            # but the referenced env file can be empty.
            "package.env": (f"app-misct/foo {os.devnull}",),
            "package.use": (f"app-misct/foo {arbitrary_package_use}",),
        }

        # Override things that may be unavailable, or may have portability
        # issues when running tests in exotic environments.
        #   prepstrip - bug #447810 (bash read builtin EINTR problem)
        true_symlinks = ("find", "prepstrip", "sed", "scanelf")
        true_binary = portage.process.find_binary("true")
        self.assertEqual(true_binary is None, False, "true command not found")

        dev_null = open(os.devnull, "wb")
        playground = ResolverPlayground(ebuilds=ebuilds, user_config=user_config)
        try:
            QueryCommand._db = playground.trees
            root_config = playground.trees[playground.eroot]["root_config"]
            portdb = root_config.trees["porttree"].dbapi
            settings = portage.config(clone=playground.settings)
            if "__PORTAGE_TEST_HARDLINK_LOCKS" in os.environ:
                settings["__PORTAGE_TEST_HARDLINK_LOCKS"] = os.environ[
                    "__PORTAGE_TEST_HARDLINK_LOCKS"
                ]
                settings.backup_changes("__PORTAGE_TEST_HARDLINK_LOCKS")

            settings.features.add("noauto")
            settings.features.add("test")
            settings["PORTAGE_PYTHON"] = portage._python_interpreter
            settings["PORTAGE_QUIET"] = "1"
            settings["PYTHONDONTWRITEBYTECODE"] = os.environ.get(
                "PYTHONDONTWRITEBYTECODE", ""
            )

            fake_bin = os.path.join(settings["EPREFIX"], "bin")
            portage.util.ensure_dirs(fake_bin)
            for x in true_symlinks:
                os.symlink(true_binary, os.path.join(fake_bin, x))

            settings["__PORTAGE_TEST_PATH_OVERRIDE"] = fake_bin
            settings.backup_changes("__PORTAGE_TEST_PATH_OVERRIDE")

            cpv = "app-misct/foo-1"
            metadata = dict(
                zip(Package.metadata_keys, portdb.aux_get(cpv, Package.metadata_keys))
            )

            pkg = Package(
                built=False,
                cpv=cpv,
                installed=False,
                metadata=metadata,
                root_config=root_config,
                type_name="ebuild",
            )
            settings.setcpv(pkg)

            # Demonstrate that settings.configdict["pkg"]["USE"] contains our arbitrary
            # package.use setting in order to trigger bug 675748.
            self.assertEqual(settings.configdict["pkg"]["USE"], arbitrary_package_use)

            # Try to trigger the config.environ() split_LC_ALL assertion for bug 925863.
            settings["LC_ALL"] = "C"

            source_ebuildpath = portdb.findname(cpv)
            self.assertNotEqual(source_ebuildpath, None)

            for phase, tree, ebuildpath in (
                ("info", "porttree", source_ebuildpath),
                ("nofetch", "porttree", source_ebuildpath),
                ("pretend", "porttree", source_ebuildpath),
                ("setup", "porttree", source_ebuildpath),
                ("unpack", "porttree", source_ebuildpath),
                ("prepare", "porttree", source_ebuildpath),
                ("configure", "porttree", source_ebuildpath),
                ("compile", "porttree", source_ebuildpath),
                ("test", "porttree", source_ebuildpath),
                ("install", "porttree", source_ebuildpath),
                ("qmerge", "porttree", source_ebuildpath),
                ("clean", "porttree", source_ebuildpath),
                ("merge", "porttree", source_ebuildpath),
                ("clean", "porttree", source_ebuildpath),
                ("config", "vartree", root_config.trees["vartree"].dbapi.findname(cpv)),
            ):
                if ebuildpath is not source_ebuildpath:
                    self.assertNotEqual(ebuildpath, None)

                pr, pw = multiprocessing.Pipe(duplex=False)

                producer = ForkProcess(
                    target=self._doebuild,
                    fd_pipes={
                        1: dev_null.fileno(),
                    },
                    args=(QueryCommand._db, pw, ebuildpath, phase),
                    kwargs={
                        "settings": settings,
                        "mydbapi": root_config.trees[tree].dbapi,
                        "tree": tree,
                        "vartree": root_config.trees["vartree"],
                        "prev_mtimes": {},
                    },
                )

                consumer = PipeReader(input_files={"producer": pr})

                task_scheduler = TaskScheduler(iter([producer, consumer]), max_jobs=2)

                try:
                    task_scheduler.start()
                finally:
                    # PipeReader closes pr
                    pw.close()

                task_scheduler.wait()
                output = portage._unicode_decode(consumer.getvalue()).rstrip("\n")

                if task_scheduler.returncode != os.EX_OK:
                    portage.writemsg(output, noiselevel=-1)

                self.assertEqual(task_scheduler.returncode, os.EX_OK)

                if phase not in ("clean", "merge", "qmerge"):
                    self.assertEqual(phase, output)

        finally:
            dev_null.close()
            playground.cleanup()
            QueryCommand._db = None

    @staticmethod
    def _doebuild(db, pw, *args, **kwargs):
        QueryCommand._db = db
        kwargs["fd_pipes"] = {
            DoebuildFdPipesTestCase.output_fd: pw.fileno(),
        }
        return portage.doebuild(*args, **kwargs)



# Copyright 2010-2024 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class EAPITestCase(TestCase):
    def testEAPI(self):
        ebuilds = {
            # EAPI-1: IUSE-defaults
            "dev-libs/A-1.0": {"EAPI": 0, "IUSE": "+foo"},
            "dev-libs/A-1.1": {"EAPI": 1, "IUSE": "+foo"},
            "dev-libs/A-1.2": {"EAPI": 2, "IUSE": "+foo"},
            "dev-libs/A-1.3": {"EAPI": 3, "IUSE": "+foo"},
            "dev-libs/A-1.4": {"EAPI": "4", "IUSE": "+foo"},
            # EAPI-1: slot deps
            "dev-libs/A-2.0": {"EAPI": 0, "DEPEND": "dev-libs/B:0"},
            "dev-libs/A-2.1": {"EAPI": 1, "DEPEND": "dev-libs/B:0"},
            "dev-libs/A-2.2": {"EAPI": 2, "DEPEND": "dev-libs/B:0"},
            "dev-libs/A-2.3": {"EAPI": 3, "DEPEND": "dev-libs/B:0"},
            "dev-libs/A-2.4": {"EAPI": "4", "DEPEND": "dev-libs/B:0"},
            # EAPI-2: use deps
            "dev-libs/A-3.0": {"EAPI": 0, "DEPEND": "dev-libs/B[foo]"},
            "dev-libs/A-3.1": {"EAPI": 1, "DEPEND": "dev-libs/B[foo]"},
            "dev-libs/A-3.2": {"EAPI": 2, "DEPEND": "dev-libs/B[foo]"},
            "dev-libs/A-3.3": {"EAPI": 3, "DEPEND": "dev-libs/B[foo]"},
            "dev-libs/A-3.4": {"EAPI": "4", "DEPEND": "dev-libs/B[foo]"},
            # EAPI-2: strong blocks
            "dev-libs/A-4.0": {"EAPI": 0, "DEPEND": "!!dev-libs/B"},
            "dev-libs/A-4.1": {"EAPI": 1, "DEPEND": "!!dev-libs/B"},
            "dev-libs/A-4.2": {"EAPI": 2, "DEPEND": "!!dev-libs/B"},
            "dev-libs/A-4.3": {"EAPI": 3, "DEPEND": "!!dev-libs/B"},
            "dev-libs/A-4.4": {"EAPI": "4", "DEPEND": "!!dev-libs/B"},
            # EAPI-4: slot operator deps
            # ~ "dev-libs/A-5.0": { "EAPI": 0, "DEPEND": "dev-libs/B:*" },
            # ~ "dev-libs/A-5.1": { "EAPI": 1, "DEPEND": "dev-libs/B:*" },
            # ~ "dev-libs/A-5.2": { "EAPI": 2, "DEPEND": "dev-libs/B:*" },
            # ~ "dev-libs/A-5.3": { "EAPI": 3, "DEPEND": "dev-libs/B:*" },
            # ~ "dev-libs/A-5.4": { "EAPI": "4", "DEPEND": "dev-libs/B:*" },
            # EAPI-4: use dep defaults
            "dev-libs/A-6.0": {"EAPI": 0, "DEPEND": "dev-libs/B[bar(+)]"},
            "dev-libs/A-6.1": {"EAPI": 1, "DEPEND": "dev-libs/B[bar(+)]"},
            "dev-libs/A-6.2": {"EAPI": 2, "DEPEND": "dev-libs/B[bar(+)]"},
            "dev-libs/A-6.3": {"EAPI": 3, "DEPEND": "dev-libs/B[bar(+)]"},
            "dev-libs/A-6.4": {"EAPI": "4", "DEPEND": "dev-libs/B[bar(+)]"},
            # EAPI-4: REQUIRED_USE
            "dev-libs/A-7.0": {
                "EAPI": 0,
                "IUSE": "foo bar",
                "REQUIRED_USE": "|| ( foo bar )",
            },
            "dev-libs/A-7.1": {
                "EAPI": 1,
                "IUSE": "foo +bar",
                "REQUIRED_USE": "|| ( foo bar )",
            },
            "dev-libs/A-7.2": {
                "EAPI": 2,
                "IUSE": "foo +bar",
                "REQUIRED_USE": "|| ( foo bar )",
            },
            "dev-libs/A-7.3": {
                "EAPI": 3,
                "IUSE": "foo +bar",
                "REQUIRED_USE": "|| ( foo bar )",
            },
            "dev-libs/A-7.4": {
                "EAPI": "4",
                "IUSE": "foo +bar",
                "REQUIRED_USE": "|| ( foo bar )",
            },
            "dev-libs/B-1": {"EAPI": 1, "IUSE": "+foo"},
            # EAPI-7: implicit || ( ) no longer satisfies deps
            "dev-libs/C-1": {
                "EAPI": "6",
                "IUSE": "foo",
                "RDEPEND": "|| ( foo? ( dev-libs/B ) )",
            },
            "dev-libs/C-2": {
                "EAPI": "7",
                "IUSE": "foo",
                "RDEPEND": "|| ( foo? ( dev-libs/B ) )",
            },
        }

        test_cases = (
            ResolverPlaygroundTestCase(["=dev-libs/A-1.0"], success=False),
            ResolverPlaygroundTestCase(
                ["=dev-libs/A-1.1"], success=True, mergelist=["dev-libs/A-1.1"]
            ),
            ResolverPlaygroundTestCase(
                ["=dev-libs/A-1.2"], success=True, mergelist=["dev-libs/A-1.2"]
            ),
            ResolverPlaygroundTestCase(
                ["=dev-libs/A-1.3"], success=True, mergelist=["dev-libs/A-1.3"]
            ),
            ResolverPlaygroundTestCase(
                ["=dev-libs/A-1.4"], success=True, mergelist=["dev-libs/A-1.4"]
            ),
            ResolverPlaygroundTestCase(["=dev-libs/A-2.0"], success=False),
            ResolverPlaygroundTestCase(
                ["=dev-libs/A-2.1"],
                success=True,
                mergelist=["dev-libs/B-1", "dev-libs/A-2.1"],
            ),
            ResolverPlaygroundTestCase(
                ["=dev-libs/A-2.2"],
                success=True,
                mergelist=["dev-libs/B-1", "dev-libs/A-2.2"],
            ),
            ResolverPlaygroundTestCase(
                ["=dev-libs/A-2.3"],
                success=True,
                mergelist=["dev-libs/B-1", "dev-libs/A-2.3"],
            ),
            ResolverPlaygroundTestCase(
                ["=dev-libs/A-2.4"],
                success=True,
                mergelist=["dev-libs/B-1", "dev-libs/A-2.4"],
            ),
            ResolverPlaygroundTestCase(["=dev-libs/A-3.0"], success=False),
            ResolverPlaygroundTestCase(["=dev-libs/A-3.1"], success=False),
            ResolverPlaygroundTestCase(
                ["=dev-libs/A-3.2"],
                success=True,
                mergelist=["dev-libs/B-1", "dev-libs/A-3.2"],
            ),
            ResolverPlaygroundTestCase(
                ["=dev-libs/A-3.3"],
                success=True,
                mergelist=["dev-libs/B-1", "dev-libs/A-3.3"],
            ),
            ResolverPlaygroundTestCase(
                ["=dev-libs/A-3.4"],
                success=True,
                mergelist=["dev-libs/B-1", "dev-libs/A-3.4"],
            ),
            ResolverPlaygroundTestCase(["=dev-libs/A-4.0"], success=False),
            ResolverPlaygroundTestCase(["=dev-libs/A-4.1"], success=False),
            ResolverPlaygroundTestCase(
                ["=dev-libs/A-4.2"], success=True, mergelist=["dev-libs/A-4.2"]
            ),
            ResolverPlaygroundTestCase(
                ["=dev-libs/A-4.3"], success=True, mergelist=["dev-libs/A-4.3"]
            ),
            ResolverPlaygroundTestCase(
                ["=dev-libs/A-4.4"], success=True, mergelist=["dev-libs/A-4.4"]
            ),
            ResolverPlaygroundTestCase(["=dev-libs/A-5.0"], success=False),
            ResolverPlaygroundTestCase(["=dev-libs/A-5.1"], success=False),
            ResolverPlaygroundTestCase(["=dev-libs/A-5.2"], success=False),
            ResolverPlaygroundTestCase(["=dev-libs/A-5.3"], success=False),
            # not implemented: EAPI-4: slot operator deps
            # ~ ResolverPlaygroundTestCase(["=dev-libs/A-5.4"], success = True, mergelist = ["dev-libs/B-1", "dev-libs/A-5.4"]),
            ResolverPlaygroundTestCase(["=dev-libs/A-6.0"], success=False),
            ResolverPlaygroundTestCase(["=dev-libs/A-6.1"], success=False),
            ResolverPlaygroundTestCase(["=dev-libs/A-6.2"], success=False),
            ResolverPlaygroundTestCase(["=dev-libs/A-6.3"], success=False),
            ResolverPlaygroundTestCase(
                ["=dev-libs/A-6.4"],
                success=True,
                mergelist=["dev-libs/B-1", "dev-libs/A-6.4"],
            ),
            ResolverPlaygroundTestCase(["=dev-libs/A-7.0"], success=False),
            ResolverPlaygroundTestCase(["=dev-libs/A-7.1"], success=False),
            ResolverPlaygroundTestCase(["=dev-libs/A-7.2"], success=False),
            ResolverPlaygroundTestCase(["=dev-libs/A-7.3"], success=False),
            ResolverPlaygroundTestCase(
                ["=dev-libs/A-7.4"], success=True, mergelist=["dev-libs/A-7.4"]
            ),
            ResolverPlaygroundTestCase(
                ["=dev-libs/C-1"], success=True, mergelist=["dev-libs/C-1"]
            ),
            ResolverPlaygroundTestCase(["=dev-libs/C-2"], success=False),
        )

        playground = ResolverPlayground(ebuilds=ebuilds)
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()

    def testBdepend(self):
        ebuilds = {
            "dev-libs/A-1.0": {"EAPI": 7},
            "dev-libs/B-1.0": {"EAPI": 7, "BDEPEND": "dev-libs/A"},
        }

        # Verify that BDEPEND is considered at all.
        test_case = ResolverPlaygroundTestCase(
            ["=dev-libs/B-1.0"],
            success=True,
            mergelist=["dev-libs/A-1.0", "dev-libs/B-1.0"],
        )

        playground = ResolverPlayground(ebuilds=ebuilds)
        try:
            playground.run_TestCase(test_case)
            self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()

    def testIdepend(self):
        ebuilds = {
            "dev-libs/A-1.0": {"EAPI": 8},
            "dev-libs/B-1.0": {"EAPI": 8, "IDEPEND": "dev-libs/A"},
        }

        test_cases = (
            # Verify that IDEPEND is considered at all.
            ResolverPlaygroundTestCase(
                ["=dev-libs/B-1.0"],
                success=True,
                mergelist=["dev-libs/A-1.0", "dev-libs/B-1.0"],
            ),
        )

        playground = ResolverPlayground(ebuilds=ebuilds)
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()
